# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# pylint: disable=C0116,W0212,R1720,C0103,C0114
#
# Note: the precision is different on V100, H100 even with the same code.
# The thresholds were adjusted on H100 as the precision seems lower on this machine.

import itertools
import unittest
import warnings

import numpy as np
import parameterized
from numpy.testing import assert_allclose
from onnx import TensorProto
from onnx.checker import check_model
from onnx.defs import onnx_opset_version
from onnx.helper import make_graph, make_model, make_node, make_opsetid, make_tensor_value_info
from onnx.numpy_helper import from_array

from onnxruntime import InferenceSession, get_available_providers

available_providers = list(get_available_providers())


class TestFloat8Gemm8(unittest.TestCase):
    def get_model_gemm(
        self,
        a_float_name="FLOAT",
        b_float_name="FLOAT",
        c_float_name="FLOAT",
        alpha=1.0,
        beta=0.0,
        transA=0,
        transB=0,
        scaleA=True,
        scaleB=True,
        scaleY=True,
        domain="",
        dtype=TensorProto.FLOAT,
        activation="NONE",
    ):
        a_proto_type = getattr(TensorProto, a_float_name)
        b_proto_type = getattr(TensorProto, b_float_name)
        c_proto_type = getattr(TensorProto, c_float_name)

        f8_set = {TensorProto.FLOAT8E4M3FN, TensorProto.FLOAT8E5M2}
        use_f8 = len({a_proto_type, b_proto_type, c_proto_type}.intersection(f8_set)) > 0

        a = make_tensor_value_info("A", TensorProto.FLOAT, [None, None])
        b = make_tensor_value_info("B", TensorProto.FLOAT, [None, None])
        d = make_tensor_value_info("Y", TensorProto.FLOAT, [None, None])

        inits = []
        kwargs = {}
        node_inputs = ["Af", "Bf"]
        inputs = [a, b]
        bias = beta != 0
        if bias:
            inputs.append(make_tensor_value_info("C", TensorProto.FLOAT, [None, None]))
            node_inputs = ["Af", "Bf", "Cf"]
            if use_f8:
                node_inputs.append("one" if scaleA else "")
                node_inputs.append("one" if scaleB else "")
                node_inputs.append("one" if scaleY else "")
        elif use_f8:
            node_inputs.append("")
            node_inputs.append("one" if scaleA else "")
            node_inputs.append("one" if scaleB else "")
            node_inputs.append("one" if scaleY else "")

        if use_f8:
            assert domain == "com.microsoft"
            inits.append(from_array(np.array([1], dtype=np.float32), name="one"))
            kwargs = {
                "domain": domain,
                "dtype": dtype,
            }
            if activation is not None:
                kwargs["activation"] = activation
            op_name = "GemmFloat8"
        elif domain == "com.microsoft":
            op_name = "GemmFloat8"
            kwargs = {
                "domain": domain,
                "dtype": dtype,
            }
        else:
            op_name = "Gemm"
        nodes = [
            make_node("Cast", ["A"], ["Af"], to=a_proto_type),
            make_node("Cast", ["B"], ["Bf"], to=b_proto_type),
            make_node("Cast", ["C"], ["Cf"], to=c_proto_type) if bias else None,
            make_node(
                op_name,
                node_inputs,
                ["Yf"],
                transA=transA,
                transB=transB,
                alpha=alpha,
                beta=beta,
                **kwargs,
            ),
            make_node("Cast", ["Yf"], ["Y"], to=TensorProto.FLOAT),
        ]
        nodes = [n for n in nodes if n is not None]
        graph = make_graph(nodes, "gemm", inputs, [d], inits)
        opset_imports = [make_opsetid("", onnx_opset_version() - 1)]
        if domain == "com.microsoft":
            opset_imports.append(make_opsetid("com.microsoft", 1))
        onnx_model = make_model(graph, opset_imports=opset_imports, ir_version=9)
        if domain != "com.microsoft":
            check_model(onnx_model)
        return onnx_model

    def common_test_model_gemm(
        self,
        a_float_name="FLOAT",
        b_float_name="FLOAT",
        c_float_name="FLOAT",
        mul=0.33,
        atol=0,
        rtol=0,
        square=True,
        **kwargs,
    ):
        if square:
            a = (np.arange(256) * 0.01).astype(np.float32).reshape((-1, 16))
            b = (np.arange(256) * -0.01).astype(np.float32).reshape((-1, 16))
            c = (np.arange(256) * 0.03).astype(np.float32).reshape((-1, 16))
            b[:, 0] += 1
        else:
            a = (np.arange(256) / 256).astype(np.float32).reshape((32, -1))
            b = (np.arange(512) / 512).astype(np.float32).reshape((32, -1))
            c = (np.arange(128) / 128).astype(np.float32).reshape((8, 16))

        feeds = {"A": a, "B": b}

        providers = ["CPUExecutionProvider"]
        if "CUDAExecutionProvider" in available_providers:
            providers = ["CUDAExecutionProvider", "CPUExecutionProvider"]
        elif "ROCMExecutionProvider" in available_providers:
            providers = [
                ("ROCMExecutionProvider", {"tunable_op_enable": "1", "tunable_op_tuning_enable": "1"}),
                ("CPUExecutionProvider", {}),
            ]

        expected = (a.T if kwargs.get("transA", 0) else a) @ (b.T if kwargs.get("transB", 0) else b)
        expected *= kwargs.get("alpha", 1.0)
        if kwargs.get("beta", 0) != 0:
            expected += kwargs["beta"] * c
            feeds["C"] = c

        onnx_model = self.get_model_gemm(**kwargs)

        ref = InferenceSession(onnx_model.SerializeToString(), providers=providers)
        y = ref.run(None, feeds)[0]
        if (
            "CUDAExecutionProvider" in providers
            and a_float_name in ("FLOAT", "FLOAT16")
            and b_float_name in ("FLOAT", "FLOAT16")
            and c_float_name in ("FLOAT", "FLOAT16")
        ):
            try:
                assert_allclose(expected, y, atol=atol, rtol=rtol)
            except Exception as e:

                def check(f):
                    try:
                        return f()[:2, :2]
                    except Exception as e:
                        return str(e)

                raise AssertionError(
                    f"Gemm ERROR len(inputs)={len(feeds)}"
                    f"\na@b=\n{check(lambda: a @ b)}"
                    f"\na.T@b=\n{check(lambda: a.T @ b)}"
                    f"\na@b.T=\n{check(lambda: a @ b.T)}"
                    f"\na.T@b.T=\n{check(lambda: a.T @ b.T)}"
                    f"\n----\nb@a=\n{check(lambda: b @ a)}"
                    f"\nb.T@a=\n{check(lambda: b.T @ a)}"
                    f"\nb@a.T=\n{check(lambda: b @ a.T)}"
                    f"\nb.T@a.T=\n{check(lambda: b.T @ a.T)}"
                    f"\n----\nexpected=\n{expected[:2, :2]}"
                    f"\n----\ngot=\n{y[:2, :2]}"
                    f"\nkwargs={kwargs}"
                ) from e

            self.assertEqual(expected.shape, y.shape)
            self.assertEqual(expected.dtype, y.dtype)

        onnx_model_f8 = self.get_model_gemm(
            a_float_name=a_float_name,
            b_float_name=b_float_name,
            c_float_name=c_float_name,
            domain="com.microsoft",
            **kwargs,
        )
        try:
            ref8 = InferenceSession(onnx_model_f8.SerializeToString(), providers=providers)
        except Exception as e:
            if "CUDA < 12.0 does not support bias" in str(e):
                return
            raise AssertionError(f"Could not load model {onnx_model_f8}") from e
        try:
            y = ref8.run(None, feeds)[0]
        except Exception as e:
            if "CUBLAS_STATUS_NOT_SUPPORTED" in str(e):
                # Skipping. This machine does not support float8.
                warnings.warn("unable to test with float8 on this machine.")
                return
            if "CK is required to support GemmFloat8 computing" in str(e):
                warnings.warn("unable to test with float8 on this build.")
                return
            raise AssertionError(f"Could not execute model {onnx_model_f8}") from e
        try:
            assert_allclose(expected, y, atol=atol, rtol=rtol)
        except Exception as e:

            def check(f):
                try:
                    return f()[:2, :2]
                except Exception as e:
                    return str(e)

            raise AssertionError(
                f"Gemm ERROR len(inputs)={len(feeds)}"
                f"\na@b=\n{check(lambda: a @ b)}"
                f"\na.T@b=\n{check(lambda: a.T @ b)}"
                f"\na@b.T=\n{check(lambda: a @ b.T)}"
                f"\na.T@b.T=\n{check(lambda: a.T @ b.T)}"
                f"\n----\nb@a=\n{check(lambda: b @ a)}"
                f"\nb.T@a=\n{check(lambda: b.T @ a)}"
                f"\nb@a.T=\n{check(lambda: b @ a.T)}"
                f"\nb.T@a.T=\n{check(lambda: b.T @ a.T)}"
                f"\n----\nexpected=\n{expected[:2, :2]}"
                f"\n----\ngot=\n{y[:2, :2]}"
                f"\nkwargs={kwargs}"
            ) from e
        self.assertEqual(expected.shape, y.shape)
        self.assertEqual(expected.dtype, y.dtype)

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float(self):
        self.common_test_model_gemm(transA=1, rtol=1e-3)

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float_default_values(self):
        self.common_test_model_gemm(transA=1, rtol=1e-3, activation=None)

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float_relu(self):
        self.common_test_model_gemm(transA=1, rtol=1e-3, activation="RELU")

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float_gelu(self):
        self.common_test_model_gemm(transA=1, rtol=1e-3, activation="GELU")

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float_bias(self):
        self.common_test_model_gemm(transA=1, beta=1.0, rtol=1e-3)

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_model_gemm_float16(self):
        self.common_test_model_gemm(
            a_float_name="FLOAT16",
            b_float_name="FLOAT16",
            c_float_name="FLOAT16",
            rtol=1e-2,
            dtype=TensorProto.FLOAT16,
            transB=1,
        )

    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    @unittest.skipIf(not hasattr(TensorProto, "FLOAT8E4M3FN"), reason="needs onnx>=1.14.0")
    def test_model_gemm_float8_e4m3(self):
        self.common_test_model_gemm(
            a_float_name="FLOAT8E4M3FN",
            b_float_name="FLOAT8E4M3FN",
            c_float_name="FLOAT8E4M3FN",
            rtol=0.5,
            dtype=TensorProto.FLOAT,
            transA=0,
            transB=1,
            alpha=10.0,
        )

    @parameterized.parameterized.expand(list(itertools.product([0, 1], [0, 1])))
    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_combinations_square_matrices(self, transA, transB):
        self.common_test_model_gemm(transA=transA, transB=transB, rtol=1e-3)

    @parameterized.parameterized.expand(
        [
            ((2, 3), (3, 5), 0, 0),
            ((2, 3), (5, 3), 0, 1),
            ((2, 3), (5, 2), 1, 1),
            ((2, 3), (2, 5), 1, 0),
        ]
    )
    @unittest.skipIf("CUDAExecutionProvider" not in available_providers, reason="Not running without CUDA.")
    def test_combinations(self, shapeA, shapeB, transA, transB):
        model = make_model(
            make_graph(
                [
                    make_node(
                        "GemmFloat8",
                        ["A", "B"],
                        ["Y"],
                        transA=transA,
                        transB=transB,
                        domain="com.microsoft",
                    )
                ],
                "f8",
                [
                    make_tensor_value_info("A", TensorProto.FLOAT, [None, None]),
                    make_tensor_value_info("B", TensorProto.FLOAT, [None, None]),
                ],
                [make_tensor_value_info("Y", TensorProto.FLOAT, [None, None])],
            ),
            opset_imports=[make_opsetid("", 19), make_opsetid("com.microsoft", 1)],
        )

        sess = InferenceSession(model.SerializeToString(), providers=["CUDAExecutionProvider", "CPUExecutionProvider"])
        a = np.arange(np.prod(shapeA)).reshape(shapeA).astype(np.float32)
        b = np.arange(np.prod(shapeB)).reshape(shapeB).astype(np.float32)
        try:
            expected = (a.T if transA else a) @ (b.T if transB else b)
        except Exception as e:
            raise AssertionError(
                f"Unable to multiply shapes={shapeA}x{shapeB}, transA={transA}, transB={transB}"
            ) from e
        try:
            got = sess.run(None, {"A": a, "B": b})
        except Exception as e:
            raise AssertionError(
                f"Unable to run Gemm with shapes={shapeA}x{shapeB}, transA={transA}, transB={transB}"
            ) from e
        self.assertEqual(expected.shape, got[0].shape)
        self.assertEqual(expected.dtype, got[0].dtype)
        assert_allclose(expected, got[0])

    @parameterized.parameterized.expand(
        [
            ("FLOAT8E4M3FN", "FLOAT16", 0, 0),
            ("FLOAT16", "FLOAT8E4M3FN", 0, 0),
            ("FLOAT16", "FLOAT8E4M3FN", 0, 1),
        ]
    )
    @unittest.skipIf("ROCMExecutionProvider" not in available_providers, reason="Not running without ROCm.")
    @unittest.skipIf(not hasattr(TensorProto, "FLOAT8E4M3FN"), reason="needs onnx>=1.14.0")
    def test_model_rocm_gemm_float8_e4m3(self, a_float_name, b_float_name, transA, transB):
        self.common_test_model_gemm(
            a_float_name=a_float_name,
            b_float_name=b_float_name,
            c_float_name="FLOAT8E4M3FN",
            rtol=0.5,
            dtype=TensorProto.FLOAT16,
            transA=0,
            transB=transB,
            scaleY=False,
            alpha=10.0,
            beta=0.0,
        )


if __name__ == "__main__":
    # TestFloat8Gemm8().test_model_gemm_float()
    unittest.main(verbosity=2)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
"""
This file controls the Ops supported by Triton codegen.
The Triton fusion on backend will try to extract subgraphs from the ONNX model which contain connected supported ops.
For each supported op, it has the following attributes:
  - domain: The domain of the op. If not specified, the op is from ONNX domain, which is empty string.
  - versions: The supported opset versions.
  - is_no_op: If the op is no-op. If not specified, the op is not no-op.
  - conditions: define some conditions to control the fusion. For example, for Ops that has axis attribute,
                a condition "axis": "-1" means only Ops that reduce on last dimension will be fused.
                For Reduce* Ops, the "axes" condition can be list of ints, such as "axes": "[-1]",
                or "axes": "single", means only support Ops reduce on single constant dimension,
                or "axes": "constant", means the axes attribute or input must be constant.
  - ignore_min_nodes: by default is False. If set to True, the fusion will ignore the min_nodes check for this Op.
                      For example, if the min_nodes is 2, the fusion will only fuse the subgraphs with 2 or more
                      non-no-op nodes. But if the ignore_min_nodes is True for ReduceSum, it's OK to fuse a single
                      ReduceSum node to the subgraph.
"""

from onnx import NodeProto

_ELEMENTWISE_OPS = {
    "Add": {"versions": [13, 14]},
    "Sub": {"versions": [13, 14]},
    "Mul": {"versions": [13, 14]},
    "Div": {"versions": [13, 14]},
    "Pow": {"versions": [13, 15]},
    "Sqrt": {"versions": [13]},
    "Exp": {"versions": [13]},
    "Where": {"versions": [9, 16]},
    "Cast": {"versions": [13]},
    "Dropout": {"versions": [13]},
    "DropoutGrad": {"domain": "com.microsoft", "versions": [1]},
    "Identity": {"versions": [13], "is_no_op": True},
    "Sum": {"versions": [13]},
    "Gelu": {"domain": "com.microsoft", "versions": [1]},
    "QuickGelu": {"domain": "com.microsoft", "versions": [1]},
    "GeluGrad": {"domain": "com.microsoft", "versions": [1]},
    "QuickGeluGrad": {"domain": "com.microsoft", "versions": [1]},
}

_REDUCTION_OPS = {
    "ReduceMean": {"versions": [13], "conditions": {"axes": "[-1]"}},
    "ReduceSum": {"versions": [13], "conditions": {"axes": "[-1]"}},
    "ReduceMax": {"versions": [13], "conditions": {"axes": "[-1]"}},
    "ReduceMin": {"versions": [13], "conditions": {"axes": "[-1]"}},
    "Softmax": {"versions": [13]},
    "SoftmaxGrad_13": {"domain": "com.microsoft", "versions": [1]},
    # Disable LayerNormalization fusion for now as it's generated Triton code is inefficient compared to C++ kernel.
    # "LayerNormalization": {"versions": [1]},
    # "LayerNormalizationGrad": {"domain": "com.microsoft", "versions": [1]},
}


def is_elementwise_node(node: NodeProto) -> bool:
    return node.op_type in _ELEMENTWISE_OPS


def is_reduction_node(node: NodeProto) -> bool:
    return node.op_type in _REDUCTION_OPS


def get_supported_ops():
    return {**_ELEMENTWISE_OPS, **_REDUCTION_OPS}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

"""This file provides wrapper for native _kernel_explorer.so library and benchmark reporter for operator"""

from __future__ import annotations

import ctypes
import json
import os
import sys
from abc import abstractmethod
from argparse import Action, ArgumentParser
from collections.abc import Callable
from contextlib import contextmanager
from dataclasses import dataclass
from fnmatch import fnmatch
from functools import wraps

build_dir = os.environ.get("KERNEL_EXPLORER_BUILD_DIR", None)
if build_dir is None:
    raise ValueError("Environment variable KERNEL_EXPLORER_BUILD_DIR is required")

if not os.path.exists(build_dir):
    raise ValueError(f"KERNEL_EXPLORER_BUILD_DIR ({build_dir}) points to nonexistent path")

# onnxruntime_pybind11_state and kernel_explorer
sys.path.insert(0, build_dir)

# pylint: disable=wrong-import-position
import onnxruntime_pybind11_state  # noqa: E402

# We need to call some functions to properly initialize so pointers in the library
available_providers = onnxruntime_pybind11_state.get_available_providers()


build_dir = os.path.realpath(build_dir)
search_paths = [build_dir]

# As Kernel Explorer makes use of utility functions in ONNXRuntime, we dlopen all relevant libraries to bring required
# symbols into global namespace, so that we don't need to worry about linking.
library_files_to_load = [
    "onnxruntime_pybind11_state.so",
    "libonnxruntime_providers_shared.so",
]
_is_cuda_available = False
_is_rocm_available = False
if "CUDAExecutionProvider" in available_providers:
    library_files_to_load.append("libonnxruntime_providers_cuda.so")
    _is_cuda_available = True
if "ROCMExecutionProvider" in available_providers:
    library_files_to_load.append("libonnxruntime_providers_rocm.so")
    _is_rocm_available = True

library_to_load = []

for lib in library_files_to_load:
    for prefix in search_paths:
        path = os.path.join(prefix, lib)
        if os.path.exists(path):
            library_to_load.append(path)
            continue

        raise OSError(f"cannot found {lib}")


# use RTLD_GLOBAL to bring all symbols to global name space
_libraries = [ctypes.CDLL(lib_path, mode=ctypes.RTLD_GLOBAL) for lib_path in library_to_load]
del library_files_to_load, library_to_load

# pylint: disable=wrong-import-position, disable=unused-import
import _kernel_explorer  # noqa: E402

# pylint: disable=wrong-import-position, disable=unused-import, disable=wildcard-import
from _kernel_explorer import *  # noqa: F403, E402


@dataclass
class _KeContext:
    sort: bool = False

    pattern = "*"

    # mapping the module to dispatch to
    dispatchable: dict | None = None
    instance_dispatchable: dict | None = None  # can be filtered with pattern

    dispatch_depth = 0

    save_tuning_results: str | None = None
    return_tuning_results: bool = False


_ke_context = _KeContext()
_ke_context.dispatchable = {}
_ke_context.instance_dispatchable = {}


# Benchmark Reporter
@dataclass
class MetricBase:
    name: str
    dtype: str
    milliseconds_duration: float

    def __lt__(self, other):
        if "Tunable" in self.name or other.duration < 0:
            return True
        if "Tunable" in other.name or self.duration < 0:
            return False

        return self.duration < other.duration

    @property
    def duration(self):
        return self.milliseconds_duration * 1000

    @abstractmethod
    def report(self) -> str:
        raise NotImplementedError()


@dataclass
class ComputeMetric(MetricBase):
    FLOPs: int

    @property
    def tflops(self):
        return self.FLOPs * 1e6 / self.duration / 1e12


@dataclass
class BandwidthMetric(MetricBase):
    bytes: int

    @property
    def gbps(self):
        return self.bytes * 1e6 / self.duration / 1e9


@dataclass
class ComputeAndBandwidthMetric(ComputeMetric, BandwidthMetric):
    pass


class InstanceBenchmarkReporter:
    def __init__(self):
        self.best = float("inf")
        self.reporters = []

    def make_report(self):
        self.reporters.sort()
        for item in self.reporters:
            if not _ke_context.sort and item.milliseconds_duration > 0 and item.milliseconds_duration < self.best:
                self.best = item.milliseconds_duration
                print(item.report(), "*")
            else:
                print(item.report())
        self.reporters.clear()

    def receive(self, status):
        self.reporters.append(status)
        if not _ke_context.sort:
            self.make_report()

    def _reset_best(self):
        self.best = float("inf")


_reporter = InstanceBenchmarkReporter()


@contextmanager
def benchmark():
    _reporter._reset_best()
    try:
        yield
    finally:
        _reporter.make_report()


def report(status):
    _reporter.receive(status)


def set_ort_severity(v):
    v = int(v)
    onnxruntime_pybind11_state.set_default_logger_severity(v)
    return v


def set_ort_verbosity(v):
    v = int(v)
    onnxruntime_pybind11_state.set_default_logger_verbosity(v)
    return v


def register_common_arguments(parser: ArgumentParser):
    class SortAction(Action):
        def __init__(self, option_strings, dest, default=False, help=None):
            super().__init__(option_strings=option_strings, dest=dest, nargs=0, default=default, help=help)

        def __call__(self, parser, namespace, values, option_string=None):
            setattr(namespace, self.dest, True)
            _ke_context.sort = True

    def set_dispatch(name):
        if name in _ke_context.dispatchable:
            dispatch = _ke_context.dispatchable[name]
            _ke_context.dispatch = dispatch
            return dispatch

        if name in _ke_context.instance_dispatchable:
            msg = f"'{name}' needs an instance to dispatch, thus it is not dispatchable from commandline."
            print(msg)
            raise ValueError(msg)

        from difflib import SequenceMatcher as Matcher

        valid_names = list(_ke_context.dispatchable.keys())
        scored_names = sorted([(Matcher(None, name, a).ratio(), a) for a in valid_names], reverse=True)
        top10 = "\n    ".join([a for _, a in scored_names[:10]])
        msg = f"'{name}' is not registered for dispatch. Top 10 matches are:\n    {top10}"
        print(msg)
        raise ValueError(msg)

    def set_pattern(pattern):
        pattern = str(pattern)
        _ke_context.pattern = pattern

    def set_save_tuning_results(path):
        _ke_context.save_tuning_results = path
        return path

    group = parser.add_argument_group("kernel explorer args", "Common arguments for kernel explorer")
    group.add_argument(
        "--sort",
        action=SortAction,
        help="control the sort of ke benchmark results based on timing",
    )
    group.add_argument(
        "--ort_default_logger_severity",
        default=2,
        choices=[0, 1, 2, 3, 4],
        type=set_ort_severity,
        help="0:Verbose, 1:Info, 2:Warning, 3:Error, 4:Fatal",
    )
    group.add_argument("--ort_default_logger_verbosity", default=0, type=set_ort_verbosity)
    group.add_argument(
        "--dispatch",
        default="profile_with_args",
        help="dispatch a registered dispatchable.",
        type=set_dispatch,
    )
    group.add_argument(
        "--pattern",
        default="*",
        help="filter the register instanced dispatchables, only matched pattern will be run.",
        type=set_pattern,
    )
    group.add_argument(
        "--save_tuning_results",
        default=None,
        type=set_save_tuning_results,
        help="patch the dispatch function to save tuning results to the specified path.",
    )

    return parser


def get_argument_parser():
    parser = ArgumentParser()
    return register_common_arguments(parser)


def has_args():
    if "--help" in sys.argv or "-h" in sys.argv or "--func" in sys.argv:
        return True

    # parse the KE args group
    parser = get_argument_parser()
    _, remainder = parser.parse_known_args(sys.argv)
    return len(remainder) > 1  # the file path is always the remainder


def is_cuda_available():
    return _is_cuda_available


def is_rocm_available():
    return _is_rocm_available


def dispatchable(f: Callable | None = None, *, pattern_arg: int | None = None):
    def wrap_dispatch(f, *args, **kwargs):
        if _ke_context.dispatch_depth == 0:
            if _ke_context.save_tuning_results is not None:
                _kernel_explorer.enable_collect_tuning_results()
        _ke_context.dispatch_depth += 1
        ret = f(*args, **kwargs)
        _ke_context.dispatch_depth -= 1
        if _ke_context.dispatch_depth == 0:
            if _ke_context.save_tuning_results is not None:
                try:
                    trs = _kernel_explorer.get_collected_tuning_results()
                    with open(_ke_context.save_tuning_results, "x") as f:
                        json.dump(trs, f)
                finally:
                    pass

            if _ke_context.return_tuning_results:
                if ret is not None:
                    print(
                        f"WARNING: kernel explorer wants to override the return value of {f.__name__},",
                        "but original return value is not None!",
                    )
                    return ret
                try:
                    trs = _kernel_explorer.get_collected_tuning_results()
                    return trs
                finally:
                    pass

        return ret

    if f is None:  # Used with ke.dispatchable(...)
        assert pattern_arg is not None

        def decorator(f):
            _ke_context.instance_dispatchable[f.__name__] = f

            @wraps(f)
            def wrapper(*args, **kwargs):
                func_name = args[pattern_arg] if isinstance(args[pattern_arg], str) else args[pattern_arg].__name__
                if not fnmatch(func_name, _ke_context.pattern):
                    print(
                        f"Trying to run {func_name},",
                        f"does not match allowed function name pattern '{_ke_context.pattern}', skip...",
                    )
                    return
                return wrap_dispatch(f, *args, **kwargs)

            return wrapper

        return decorator

    else:  # Used with @ke.dispatchable
        _ke_context.dispatchable[f.__name__] = f

        @wraps(f)
        def wrapper(*args, **kwargs):
            return wrap_dispatch(f, *args, **kwargs)

        return wrapper


def set_dispatchable_pattern(p: str = "*"):
    _ke_context.pattern = p


def set_return_tuning_results(b: bool = True):
    _ke_context.return_tuning_results = b



import os

import torch

from onnxruntime.capi import _pybind_state as C
from onnxruntime.training import ortmodule
from onnxruntime.training.ortmodule.experimental.json_config import load_from_json


class Net(torch.nn.Module):
    def __init__(self, input_size=784, hidden_size=500, num_classes=10):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1):
        out = self.fc1(input1)
        out = self.relu(out)
        out = self.fc2(out)
        return out


def test_load_config_from_json_1():
    device = "cuda"
    model = ortmodule.ORTModule(Net().to(device))

    # load from json once.
    path_to_json = os.path.join(os.getcwd(), "orttraining_test_ortmodule_experimental_json_config_2.json")
    load_from_json(model, path_to_json)

    # load from json another time
    path_to_json = os.path.join(os.getcwd(), "orttraining_test_ortmodule_experimental_json_config_1.json")
    load_from_json(model, path_to_json)

    for training_mode in [True, False]:
        ort_model_attributes = model._torch_module._execution_manager(training_mode)

        # test propagate cast ops
        assert (
            ort_model_attributes._runtime_options.propagate_cast_ops_strategy == C.PropagateCastOpsStrategy.FLOOD_FILL
        )
        assert ort_model_attributes._runtime_options.propagate_cast_ops_level == 3
        assert ort_model_attributes._runtime_options.propagate_cast_ops_allow == ["ABC", "DEF"]

        # test use external gpu allocator
        assert ort_model_attributes._runtime_options.use_external_gpu_allocator is False

        # test enable custom autograd function
        assert ort_model_attributes._runtime_options.enable_custom_autograd_function is True

        # test use static shape
        assert ort_model_attributes._runtime_options.use_static_shape is True

        # test run symbolic shape inference
        assert ort_model_attributes._runtime_options.run_symbolic_shape_infer is False

        # test enable grad acc optimization
        assert ort_model_attributes._runtime_options.enable_grad_acc_optimization is True

        # test skip check
        assert ort_model_attributes._runtime_options.skip_check.value == 14

        # test debug options
        assert ort_model_attributes._debug_options.save_onnx_models.save is True
        assert ort_model_attributes._debug_options.save_onnx_models.name_prefix == "my_model"
        assert ort_model_attributes._debug_options.logging.log_level.name == "VERBOSE"

        # test use memory aware gradient builder.
        assert ort_model_attributes._runtime_options.use_memory_efficient_gradient is False

        # test fallback policy
        assert ort_model_attributes._fallback_manager.policy.value == 1

        # assert onnx opset version
        assert ort_model_attributes._runtime_options.onnx_opset_version == 13


def test_load_config_from_json_2():
    device = "cuda"
    model = ortmodule.ORTModule(Net().to(device))

    # load from json once.
    path_to_json = os.path.join(os.getcwd(), "orttraining_test_ortmodule_experimental_json_config_1.json")
    load_from_json(model, path_to_json)

    # load from json another time
    path_to_json = os.path.join(os.getcwd(), "orttraining_test_ortmodule_experimental_json_config_2.json")
    load_from_json(model, path_to_json)

    for training_mode in [True, False]:
        ort_model_attributes = model._torch_module._execution_manager(training_mode)

        # test propagate cast ops
        assert (
            ort_model_attributes._runtime_options.propagate_cast_ops_strategy
            == C.PropagateCastOpsStrategy.INSERT_AND_REDUCE
        )
        assert ort_model_attributes._runtime_options.propagate_cast_ops_level == 5
        assert ort_model_attributes._runtime_options.propagate_cast_ops_allow == ["XYZ", "PQR"]

        # test use external gpu allocator
        assert ort_model_attributes._runtime_options.use_external_gpu_allocator is True

        # test enable custom autograd function
        assert ort_model_attributes._runtime_options.enable_custom_autograd_function is False

        # test use static shape
        assert ort_model_attributes._runtime_options.use_static_shape is False

        # test run symbolic shape inference
        assert ort_model_attributes._runtime_options.run_symbolic_shape_infer is True

        # test enable grad acc optimization
        assert ort_model_attributes._runtime_options.enable_grad_acc_optimization is False

        # test skip check
        assert ort_model_attributes._runtime_options.skip_check.value == 10

        # test debug options
        assert ort_model_attributes._debug_options.save_onnx_models.save is True
        assert ort_model_attributes._debug_options.save_onnx_models.name_prefix == "my_other_model"
        assert ort_model_attributes._debug_options.logging.log_level.name == "INFO"

        # test use memory aware gradient builder.
        assert ort_model_attributes._runtime_options.use_memory_efficient_gradient is True

        # test fallback policy
        assert ort_model_attributes._fallback_manager.policy.value == 250

        # assert onnx opset version
        assert ort_model_attributes._runtime_options.onnx_opset_version == 12



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


import argparse
import datetime
import os
import sys

# ingest from dataframe
import pandas
from azure.kusto.data import DataFormat, KustoConnectionStringBuilder
from azure.kusto.ingest import IngestionProperties, QueuedIngestClient, ReportLevel


def parse_arguments():
    parser = argparse.ArgumentParser(description="ONNXRuntime binary size uploader for dashboard")
    parser.add_argument("--commit_hash", help="Full Git commit hash")
    parser.add_argument(
        "--build_project",
        default="Lotus",
        choices=["Lotus", "onnxruntime"],
        help="Lotus or onnxruntime build project, to construct the build URL",
    )
    parser.add_argument("--build_id", help="Build Id")
    parser.add_argument("--size_data_file", help="Path to file that contains the binary size data")
    parser.add_argument(
        "--ignore_db_error", action="store_true", help="Ignore database errors while executing this script"
    )

    return parser.parse_args()


# Assumes size_data_file is a csv file with a header line, containing binary sizes and other attributes
# CSV fields are:
#    os,arch,build_config,size
# No empty line or space between fields expected


def get_binary_sizes(size_data_file):
    binary_size = []
    with open(size_data_file) as f:
        line = f.readline()
        headers = line.strip().split(",")
        while line:
            line = f.readline()
            if not line:
                break
            linedata = line.strip().split(",")
            tablerow = {}
            for i in range(len(headers)):
                if headers[i] == "size":
                    tablerow[headers[i]] = int(linedata[i])
                else:
                    tablerow[headers[i]] = linedata[i]
            binary_size.append(tablerow)
    return binary_size


def write_to_db(binary_size_data, args):
    # connect to database
    cluster = "https://ingest-onnxruntimedashboarddb.southcentralus.kusto.windows.net"
    kcsb = KustoConnectionStringBuilder.with_az_cli_authentication(cluster)
    # The authentication method will be taken from the chosen KustoConnectionStringBuilder.
    client = QueuedIngestClient(kcsb)
    fields = ["build_time", "build_id", "build_project", "commit_id", "os", "arch", "build_config", "size", "Branch"]
    now_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    branch_name = os.environ.get("BUILD_SOURCEBRANCHNAME", "main")
    rows = []
    for row in binary_size_data:
        rows.append(
            [
                now_str,
                args.build_id,
                args.build_project,
                args.commit_hash,
                row["os"],
                row["arch"],
                row["build_config"],
                row["size"],
                branch_name.lower(),
            ]
        )
    ingestion_props = IngestionProperties(
        database="powerbi",
        table="binary_size",
        data_format=DataFormat.CSV,
        report_level=ReportLevel.FailuresAndSuccesses,
    )
    df = pandas.DataFrame(data=rows, columns=fields)
    client.ingest_from_dataframe(df, ingestion_properties=ingestion_props)


if __name__ == "__main__":
    args = parse_arguments()
    binary_size_data = get_binary_sizes(args.size_data_file)
    try:
        write_to_db(binary_size_data, args)
    except BaseException as e:
        print(str(e))
        # If there is DB connection error, and we choose '--ignore_db_error'
        # we can let the script exit clean in order not to fail the pipeline
        if not args.ignore_db_error:
            sys.exit(1)



import argparse
import logging
import time

import torch
from torchvision import datasets, transforms

import onnxruntime
from onnxruntime.training.ortmodule import DebugOptions, ORTModule


class NeuralNet(torch.nn.Module):
    def __init__(self, input_size, hidden_size, num_classes):
        super().__init__()

        self.fc1 = torch.nn.Linear(input_size, hidden_size)
        self.relu = torch.nn.ReLU()
        self.fc2 = torch.nn.Linear(hidden_size, num_classes)

    def forward(self, input1):
        out = self.fc1(input1)
        out = self.relu(out)
        out = self.fc2(out)
        return out


def train(args, model, device, optimizer, loss_fn, train_loader, epoch):
    print(f"\n======== Epoch {epoch + 1} / {args.epochs} with batch size {args.batch_size} ========")
    model.train()
    # Measure how long the training epoch takes.
    t0 = time.time()
    start_time = t0

    # Reset the total loss for this epoch.
    total_loss = 0

    for iteration, (data, target) in enumerate(train_loader):
        if iteration == args.train_steps:
            break
        data, target = data.to(device), target.to(device)  # noqa: PLW2901
        data = data.reshape(data.shape[0], -1)  # noqa: PLW2901

        optimizer.zero_grad()
        probability = model(data)

        if args.view_graphs:
            import torchviz

            pytorch_backward_graph = torchviz.make_dot(probability, params=dict(list(model.named_parameters())))
            pytorch_backward_graph.view()

        loss = loss_fn(probability, target)
        # Accumulate the training loss over all of the batches so that we can
        # calculate the average loss at the end. `loss` is a Tensor containing a
        # single value; the `.item()` function just returns the Python value
        # from the tensor.
        total_loss += loss.item()

        loss.backward()
        optimizer.step()

        # Stats
        if iteration % args.log_interval == 0:
            curr_time = time.time()
            elapsed_time = curr_time - start_time
            print(
                f"[{iteration * len(data):5}/{len(train_loader.dataset):5} ({100.0 * iteration / len(train_loader):2.0f}%)]\tLoss: {loss:.6f}\tExecution time: {elapsed_time:.4f}"
            )
            start_time = curr_time

    # Calculate the average loss over the training data.
    avg_train_loss = total_loss / len(train_loader)

    epoch_time = time.time() - t0
    print(f"\n  Average training loss: {avg_train_loss:.2f}")
    print(f"  Training epoch took: {epoch_time:.4f}s")
    return epoch_time


def test(args, model, device, loss_fn, test_loader):
    model.eval()

    t0 = time.time()

    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)  # noqa: PLW2901
            data = data.reshape(data.shape[0], -1)  # noqa: PLW2901
            output = model(data)

            # Stats
            test_loss += loss_fn(output, target, False).item()
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()
    test_loss /= len(test_loader.dataset)
    print(
        f"\nTest set: Batch size: {args.test_batch_size}, Average loss: {test_loss:.4f}, Accuracy: {correct}/{len(test_loader.dataset)} ({100.0 * correct / len(test_loader.dataset):.0f}%)\n"
    )

    # Report the final accuracy for this validation run.
    epoch_time = time.time() - t0
    accuracy = float(correct) / len(test_loader.dataset)
    print(f"  Accuracy: {accuracy:.2f}")
    print(f"  Validation took: {epoch_time:.4f}s")
    return epoch_time, accuracy


def my_loss(x, target, is_train=True):
    if is_train:
        return torch.nn.CrossEntropyLoss()(x, target)
    else:
        return torch.nn.CrossEntropyLoss(reduction="sum")(x, target)


def main():
    # Training settings
    parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
    parser.add_argument(
        "--train-steps",
        type=int,
        default=-1,
        metavar="N",
        help="number of steps to train. Set -1 to run through whole dataset (default: -1)",
    )
    parser.add_argument("--lr", type=float, default=0.01, metavar="LR", help="learning rate (default: 0.01)")
    parser.add_argument(
        "--batch-size", type=int, default=32, metavar="N", help="input batch size for training (default: 32)"
    )
    parser.add_argument(
        "--test-batch-size", type=int, default=64, metavar="N", help="input batch size for testing (default: 64)"
    )
    parser.add_argument("--no-cuda", action="store_true", default=False, help="disables CUDA training")
    parser.add_argument("--seed", type=int, default=42, metavar="S", help="random seed (default: 42)")
    parser.add_argument("--pytorch-only", action="store_true", default=False, help="disables ONNX Runtime training")
    parser.add_argument(
        "--log-interval",
        type=int,
        default=300,
        metavar="N",
        help="how many batches to wait before logging training status (default: 300)",
    )
    parser.add_argument("--view-graphs", action="store_true", default=False, help="views forward and backward graphs")
    parser.add_argument(
        "--export-onnx-graphs", action="store_true", default=False, help="export ONNX graphs to current directory"
    )
    parser.add_argument("--epochs", type=int, default=5, metavar="N", help="number of epochs to train (default: 10)")
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        help="Log level (default: WARNING)",
    )
    parser.add_argument("--data-dir", type=str, default="./mnist", help="Path to the mnist data directory")

    args = parser.parse_args()

    # Common setup
    torch.manual_seed(args.seed)
    onnxruntime.set_seed(args.seed)

    if not args.no_cuda and torch.cuda.is_available():
        device = "cuda"
    else:
        device = "cpu"

    ## Data loader
    train_loader = torch.utils.data.DataLoader(
        datasets.MNIST(
            args.data_dir,
            train=True,
            download=True,
            transform=transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]),
        ),
        batch_size=args.batch_size,
        shuffle=True,
    )
    test_loader = None
    if args.test_batch_size > 0:
        test_loader = torch.utils.data.DataLoader(
            datasets.MNIST(
                args.data_dir,
                train=False,
                transform=transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]),
            ),
            batch_size=args.test_batch_size,
            shuffle=True,
        )

    # Model architecture
    model = NeuralNet(input_size=784, hidden_size=500, num_classes=10).to(device)
    if not args.pytorch_only:
        print("Training MNIST on ORTModule....")

        # Just for future debugging
        debug_options = DebugOptions(save_onnx=args.export_onnx_graphs, onnx_prefix="MNIST")

        model = ORTModule(model, debug_options)

        # Set log level
        numeric_level = getattr(logging, args.log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError(f"Invalid log level: {args.log_level}")
        logging.basicConfig(level=numeric_level)
    else:
        print("Training MNIST on vanilla PyTorch....")
    optimizer = torch.optim.SGD(model.parameters(), lr=args.lr)

    # Train loop
    total_training_time, total_test_time, epoch_0_training, validation_accuracy = 0, 0, 0, 0
    for epoch in range(args.epochs):
        total_training_time += train(args, model, device, optimizer, my_loss, train_loader, epoch)
        if not args.pytorch_only and epoch == 0:
            epoch_0_training = total_training_time
        if args.test_batch_size > 0:
            test_time, validation_accuracy = test(args, model, device, my_loss, test_loader)
            total_test_time += test_time

    assert validation_accuracy > 0.92

    print("\n======== Global stats ========")
    if not args.pytorch_only:
        estimated_export = 0
        if args.epochs > 1:
            estimated_export = epoch_0_training - (total_training_time - epoch_0_training) / (args.epochs - 1)
            print(f"  Estimated ONNX export took:               {estimated_export:.4f}s")
        else:
            print("  Estimated ONNX export took:               Estimate available when epochs > 1 only")
        print(f"  Accumulated training without export took: {total_training_time - estimated_export:.4f}s")
    print(f"  Accumulated training took:                {total_training_time:.4f}s")
    print(f"  Accumulated validation took:              {total_test_time:.4f}s")


if __name__ == "__main__":
    main()



import onnx
from onnx import OperatorSetIdProto, TensorProto, helper


def GenerateModel(model_name, has_casts=False):  # noqa: N802
    nodes = [  # SimplifiedLayerNorm subgraph
        helper.make_node("Pow", ["cast_A" if has_casts else "A", "pow_in_2"], ["pow_out"], "pow"),
        helper.make_node("ReduceMean", ["pow_out"], ["rd2_out"], "reduce", axes=[-1], keepdims=1),
        helper.make_node("Add", ["rd2_out", "const_e12"], ["add1_out"], "add"),
        helper.make_node("Sqrt", ["add1_out"], ["sqrt_out"], "sqrt"),
        helper.make_node("Div", ["cast_A" if has_casts else "A", "sqrt_out"], ["div_out"], "div"),
        helper.make_node("Mul", ["gamma", "cast_div_out" if has_casts else "div_out"], ["C"], "mul"),
    ]

    if has_casts:
        nodes.extend(
            [
                helper.make_node("Cast", ["A"], ["cast_A"], "cast A", to=1),
                helper.make_node("Cast", ["div_out"], ["cast_div_out"], "cast_2", to=10),
            ]
        )

    initializers = [  # initializers
        helper.make_tensor("pow_in_2", TensorProto.FLOAT, [], [2]),
        helper.make_tensor("const_e12", TensorProto.FLOAT, [], [1e-12]),
        helper.make_tensor(
            "gamma",
            TensorProto.FLOAT16 if has_casts else TensorProto.FLOAT,
            [4],
            [1, 2, 3, 4],
        ),
    ]

    input_type = TensorProto.FLOAT16 if has_casts else TensorProto.FLOAT
    output_type = TensorProto.FLOAT16 if has_casts else TensorProto.FLOAT

    graph = helper.make_graph(
        nodes,
        "SimplifiedLayerNorm",  # name
        [  # inputs
            helper.make_tensor_value_info("A", input_type, [16, 32, 4]),
        ],
        [  # outputs
            helper.make_tensor_value_info("C", output_type, [16, 32, 4]),
        ],
        initializers,
    )

    onnxdomain = OperatorSetIdProto()
    onnxdomain.version = 12
    # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
    onnxdomain.domain = ""
    msdomain = OperatorSetIdProto()
    msdomain.version = 1
    msdomain.domain = "com.microsoft"
    opsets = [onnxdomain, msdomain]

    model = helper.make_model(graph, opset_imports=opsets)
    onnx.save(model, model_name)


GenerateModel("layer_norm_t5.onnx")
GenerateModel("simplified_layer_norm_with_casts.onnx", True)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import glob
import json
import os
import pathlib
import shutil
import subprocess
import sys

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
REPO_DIR = os.path.normpath(os.path.join(SCRIPT_DIR, "..", "..", "..", ".."))
BUILD_PY = os.path.join(REPO_DIR, "tools", "ci_build", "build.py")

# We by default will build below 3 archs
DEFAULT_BUILD_OSX_ARCHS = {
    "iphoneos": ["arm64"],
    "iphonesimulator": ["arm64", "x86_64"],
}


def _parse_build_settings(args):
    with open(args.build_settings_file.resolve()) as f:
        build_settings_data = json.load(f)

    build_settings = {}

    build_settings["build_osx_archs"] = build_settings_data.get("build_osx_archs", DEFAULT_BUILD_OSX_ARCHS)

    if "build_params" in build_settings_data:
        build_settings["build_params"] = build_settings_data["build_params"]
    else:
        raise ValueError("build_params is required in the build config file")

    return build_settings


# Build fat framework for all archs of a single sysroot
# For example, arm64 and x86_64 for iphonesimulator
def _build_for_apple_sysroot(
    build_config, intermediates_dir, base_build_command, sysroot, archs, build_dynamic_framework
):
    # paths of the onnxruntime libraries for different archs
    ort_libs = []
    info_plist_path = ""

    # Build binary for each arch, one by one
    for current_arch in archs:
        build_dir_current_arch = os.path.join(intermediates_dir, sysroot + "_" + current_arch)
        # Use MacOS SDK for Catalyst builds
        apple_sysroot = "macosx" if sysroot == "macabi" else sysroot
        build_command = [
            *base_build_command,
            "--apple_sysroot=" + apple_sysroot,
            "--osx_arch=" + current_arch,
            "--build_dir=" + build_dir_current_arch,
        ]

        # the actual build process for current arch
        subprocess.run(build_command, shell=False, check=True, cwd=REPO_DIR)

        # get the compiled lib path
        framework_dir = os.path.join(
            build_dir_current_arch,
            build_config,
            build_config + "-" + sysroot,
            (
                "onnxruntime.framework"
                if build_dynamic_framework
                else os.path.join("static_framework", "onnxruntime.framework")
            ),
        )
        ort_libs.append(os.path.join(framework_dir, "onnxruntime"))

        # We only need to copy Info.plist, framework_info.json, and headers once since they are the same
        if not info_plist_path:
            info_plist_path = os.path.join(build_dir_current_arch, build_config, "Info.plist")
            framework_info_path = os.path.join(build_dir_current_arch, build_config, "framework_info.json")
            headers = glob.glob(os.path.join(framework_dir, "Headers", "*.h"))

    # manually create the fat framework
    framework_dir = os.path.join(intermediates_dir, "frameworks", sysroot, "onnxruntime.framework")
    # remove the existing framework if any
    if os.path.exists(framework_dir):
        shutil.rmtree(framework_dir)
    pathlib.Path(framework_dir).mkdir(parents=True, exist_ok=True)

    # copy the Info.plist, framework_info.json, and header files

    # macos requires different framework structure:
    # https://developer.apple.com/library/archive/documentation/MacOSX/Conceptual/BPFrameworks/Concepts/FrameworkAnatomy.html
    if sysroot == "macosx" or sysroot == "macabi":
        # create headers and resources directory
        header_dir = os.path.join(framework_dir, "Versions", "A", "Headers")
        resource_dir = os.path.join(framework_dir, "Versions", "A", "Resources")
        pathlib.Path(header_dir).mkdir(parents=True, exist_ok=True)
        pathlib.Path(resource_dir).mkdir(parents=True, exist_ok=True)

        shutil.copy(info_plist_path, resource_dir)
        shutil.copy(framework_info_path, os.path.dirname(framework_dir))

        for _header in headers:
            shutil.copy(_header, header_dir)

        # use lipo to create a fat ort library
        lipo_command = ["lipo", "-create"]
        lipo_command += ort_libs
        lipo_command += ["-output", os.path.join(framework_dir, "Versions", "A", "onnxruntime")]
        subprocess.run(lipo_command, shell=False, check=True)

        # create the symbolic link
        pathlib.Path(os.path.join(framework_dir, "Versions", "Current")).symlink_to("A", target_is_directory=True)
        pathlib.Path(os.path.join(framework_dir, "Headers")).symlink_to(
            "Versions/Current/Headers", target_is_directory=True
        )
        pathlib.Path(os.path.join(framework_dir, "Resources")).symlink_to(
            "Versions/Current/Resources", target_is_directory=True
        )
        pathlib.Path(os.path.join(framework_dir, "onnxruntime")).symlink_to("Versions/Current/onnxruntime")

    else:
        shutil.copy(info_plist_path, framework_dir)
        shutil.copy(framework_info_path, os.path.dirname(framework_dir))
        header_dir = os.path.join(framework_dir, "Headers")
        pathlib.Path(header_dir).mkdir(parents=True, exist_ok=True)

        for _header in headers:
            shutil.copy(_header, header_dir)

        # use lipo to create a fat ort library
        lipo_command = ["lipo", "-create"]
        lipo_command += ort_libs
        lipo_command += ["-output", os.path.join(framework_dir, "onnxruntime")]
        subprocess.run(lipo_command, shell=False, check=True)

    return framework_dir


def _merge_framework_info_files(files, output_file):
    merged_data = {}

    for file in files:
        with open(file) as f:
            data = json.load(f)
            for platform, values in data.items():
                assert platform not in merged_data, f"Duplicate platform value: {platform}"
                merged_data[platform] = values

    with open(output_file, "w") as f:
        json.dump(merged_data, f, indent=2)


def _build_package(args):
    build_settings = _parse_build_settings(args)
    build_dir = os.path.abspath(args.build_dir)

    # Temp dirs to hold building results
    intermediates_dir = os.path.join(build_dir, "intermediates")
    build_config = args.config

    # build framework for individual sysroot
    framework_dirs = []
    framework_info_files_to_merge = []
    public_headers_path = ""
    for sysroot in build_settings["build_osx_archs"]:
        base_build_command = (
            [sys.executable, BUILD_PY]
            + build_settings["build_params"]["base"]
            + build_settings["build_params"][sysroot]
            + ["--config=" + build_config]
        )

        if args.include_ops_by_config is not None:
            base_build_command += ["--include_ops_by_config=" + str(args.include_ops_by_config.resolve())]

        if args.path_to_protoc_exe is not None:
            base_build_command += ["--path_to_protoc_exe=" + str(args.path_to_protoc_exe.resolve())]

        framework_dir = _build_for_apple_sysroot(
            build_config,
            intermediates_dir,
            base_build_command,
            sysroot,
            build_settings["build_osx_archs"][sysroot],
            args.build_dynamic_framework,
        )
        framework_dirs.append(framework_dir)

        curr_framework_info_path = os.path.join(os.path.dirname(framework_dir), "framework_info.json")
        framework_info_files_to_merge.append(curr_framework_info_path)

        # headers for each sysroot are the same, pick one of them
        if not public_headers_path:
            public_headers_path = os.path.join(os.path.dirname(framework_dir), "onnxruntime.framework", "Headers")

    # create the folder for xcframework and copy the LICENSE and framework_info.json file
    xcframework_dir = os.path.join(build_dir, "framework_out")
    pathlib.Path(xcframework_dir).mkdir(parents=True, exist_ok=True)
    shutil.copy(os.path.join(REPO_DIR, "LICENSE"), xcframework_dir)
    shutil.copytree(public_headers_path, os.path.join(xcframework_dir, "Headers"), dirs_exist_ok=True, symlinks=True)
    _merge_framework_info_files(framework_info_files_to_merge, os.path.join(build_dir, "xcframework_info.json"))

    # remove existing xcframework if any
    xcframework_path = os.path.join(xcframework_dir, "onnxruntime.xcframework")
    if os.path.exists(xcframework_path):
        shutil.rmtree(xcframework_path)

    # Assemble the final xcframework
    build_xcframework_cmd = ["xcrun", "xcodebuild", "-create-xcframework", "-output", xcframework_path]
    for framework_dir in framework_dirs:
        build_xcframework_cmd.extend(["-framework", framework_dir])

    subprocess.run(build_xcframework_cmd, shell=False, check=True, cwd=REPO_DIR)


def parse_args():
    parser = argparse.ArgumentParser(
        os.path.basename(__file__),
        description="""Create iOS framework and podspec for one or more osx_archs (xcframework)
        and building properties specified in the given build config file, see
        tools/ci_build/github/apple/default_full_apple_framework_build_settings.json for details.
        The output of the final xcframework and podspec can be found under [build_dir]/framework_out.
        Please note, this building script will only work on macOS.
        """,
    )

    parser.add_argument(
        "--build_dir",
        type=pathlib.Path,
        default=os.path.join(REPO_DIR, "build/apple_framework"),
        help="Provide the root directory for build output",
    )

    parser.add_argument(
        "--include_ops_by_config",
        type=pathlib.Path,
        help="Include ops from config file. See /docs/Reduced_Operator_Kernel_build.md for more information.",
    )

    parser.add_argument(
        "--config",
        type=str,
        default="Release",
        choices=["Debug", "MinSizeRel", "Release", "RelWithDebInfo"],
        help="Configuration to build.",
    )

    parser.add_argument(
        "--build_dynamic_framework",
        action="store_true",
        help="Build Dynamic Framework (default is build static framework).",
    )

    parser.add_argument(
        "build_settings_file", type=pathlib.Path, help="Provide the file contains settings for building iOS framework"
    )

    parser.add_argument("--path_to_protoc_exe", type=pathlib.Path, help="Path to protoc exe.")

    args = parser.parse_args()

    if not args.build_settings_file.resolve().is_file():
        raise FileNotFoundError(f"Build config file {args.build_settings_file.resolve()} is not a file.")

    if args.include_ops_by_config is not None:
        include_ops_by_config_file = args.include_ops_by_config.resolve()
        if not include_ops_by_config_file.is_file():
            raise FileNotFoundError(f"Include ops config file {include_ops_by_config_file} is not a file.")

    return args


def main():
    args = parse_args()
    _build_package(args)


if __name__ == "__main__":
    main()



from onnx import OperatorSetIdProto, TensorProto, helper, save

opsets = []
onnxdomain = OperatorSetIdProto()
onnxdomain.version = 19
onnxdomain.domain = ""  # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
opsets.append(onnxdomain)

msdomain = OperatorSetIdProto()
msdomain.version = 1
msdomain.domain = "com.microsoft"
opsets.append(msdomain)

ai_ml_domain = OperatorSetIdProto()
ai_ml_domain.version = 4
ai_ml_domain.domain = "ai.onnx.ml"
opsets.append(ai_ml_domain)

kwargs = {}
kwargs["opset_imports"] = opsets


def generate_model(model_name):
    # Create models with consecutive label encoders
    nodes = [  # subgraph
        # string -> int -> string
        helper.make_node(
            "LabelEncoder",
            ["A"],
            ["le_1_int_1"],
            "le_1_int_1",
            domain="ai.onnx.ml",
            keys_strings=["a", "b", "c"],
            values_int64s=[0, 1, 2],
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_1_int_1"],
            ["le_1_string_2"],
            "le_1_string_2",
            domain="ai.onnx.ml",
            keys_int64s=[2, 1, 0],
            values_strings=["a", "b", "c"],
            default_string="default",
        ),
        # string -> string -> string
        helper.make_node(
            "LabelEncoder",
            ["A"],
            ["le_2_string_1"],
            "le_2_string_1",
            domain="ai.onnx.ml",
            keys_strings=["a", "b", "c"],
            values_strings=["C", "B", "A"],
            default_string="D",
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_2_string_1"],
            ["le_2_string_2"],
            "le_2_string_2",
            domain="ai.onnx.ml",
            keys_strings=["A", "B", "C", "D"],
            values_strings=["a", "b", "c", "d"],
            default_string="default",
        ),
        # string -> string -> int -> string
        helper.make_node(
            "LabelEncoder",
            ["A"],
            ["le_3_string_1"],
            "le_3_string_1",
            domain="ai.onnx.ml",
            keys_strings=["a", "b", "c"],
            values_strings=["C", "B", "A"],
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_3_string_1"],
            ["le_3_int_2"],
            "le_3_int_2",
            domain="ai.onnx.ml",
            keys_strings=["A", "B", "C"],
            values_int64s=[1, 2, 3],
            default_int64=-1,
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_3_int_2"],
            ["le_3_string_3"],
            "le_3_string_3",
            domain="ai.onnx.ml",
            keys_int64s=[1, 2, 3],
            values_strings=["a", "b", "c"],
            default_string="d",
        ),
        # middle encoder is graph output
        helper.make_node(
            "LabelEncoder",
            ["A"],
            ["le_4_int_1"],
            "le_4_int_1",
            domain="ai.onnx.ml",
            keys_strings=["a", "b", "c"],
            values_int64s=[0, 1, 2],
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_4_int_1"],
            ["le_4_string_2"],
            "le_4_string_2",
            domain="ai.onnx.ml",
            keys_int64s=[0, 1, 2],
            values_strings=["a", "b", "c"],
        ),
        helper.make_node("Identity", ["le_4_int_1"], ["Y"], "output"),
        # middle encoder is consumed twice
        helper.make_node(
            "LabelEncoder",
            ["A"],
            ["le_5_int_1"],
            "le_5_int_1",
            domain="ai.onnx.ml",
            keys_strings=["a", "b", "c"],
            values_int64s=[0, 1, 2],
        ),
        helper.make_node(
            "LabelEncoder",
            ["le_5_int_1"],
            ["le_5_string_2"],
            "le_5_string_2",
            domain="ai.onnx.ml",
            keys_int64s=[0, 1, 2],
            values_strings=["a", "b", "c"],
        ),
        helper.make_node("Mul", ["le_5_int_1", "le_5_int_1"], ["mul_5"], "mul_5"),
    ]

    inputs = [  # inputs
        helper.make_tensor_value_info("A", TensorProto.STRING, ["N"]),
    ]

    graph = helper.make_graph(
        nodes,
        "LabelEncoder",  # name
        inputs,
        [  # outputs
            helper.make_tensor_value_info("le_1_string_2", TensorProto.STRING, ["N"]),
            helper.make_tensor_value_info("le_2_string_2", TensorProto.STRING, ["N"]),
            helper.make_tensor_value_info("le_3_string_3", TensorProto.STRING, ["N"]),
            helper.make_tensor_value_info("le_4_string_2", TensorProto.STRING, ["N"]),
            helper.make_tensor_value_info("Y", TensorProto.INT64, ["N"]),
            helper.make_tensor_value_info("mul_5", TensorProto.INT64, ["N"]),
        ],
        [],
    )

    model = helper.make_model(graph, **kwargs)
    save(model, model_name)


if __name__ == "__main__":
    generate_model("label_encoder.onnx")



import logging

import numpy as np
import torch
from torchvision import datasets, transforms

import onnxruntime.quantization as quantization


def preprocess(input_model_dir, output_model_dir):
    """Preprocesses the given onnx model for quantization. Unused for QAT process."""
    quantization.shape_inference.quant_pre_process(input_model_dir, output_model_dir)


class MnistDataReader:
    """Generates calibration data from the MNIST dataset."""

    def __init__(self, data_dir, num_samples):
        self.data_dir = data_dir
        self.num_samples = num_samples

        mnist_data = datasets.MNIST(
            root=data_dir,
            train=True,
            download=True,
            transform=transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]),
        )
        probs = torch.tensor([1 / mnist_data.data.shape[0]] * num_samples, dtype=torch.float32)
        index = probs.multinomial(num_samples=num_samples, replacement=False)
        self.data = torch.reshape(mnist_data.data[index], [num_samples, -1])
        self.input_name = "input-0"
        self.datasize = num_samples
        self.data_loader = iter(self.data)

    def get_next(self):
        next_array = next(self.data_loader, None)
        if next_array is None:
            return None
        return {self.input_name: np.expand_dims(next_array, axis=0).tolist()}


def quantize_static(input_model_dir, output_model_dir):
    """Statically quantizes the given onnx model using onnxruntime's static quantization tools.

    Calibration data is generated from the MNIST dataset.
    """

    extra_options = {"AddQDQPairToWeight": True, "QuantizeBias": False}
    logging.info("Using calibration data from MNIST dataset picking 1000 samples at random.")
    calibration_data_reader = MnistDataReader("data", 1000)

    # Quantize the model
    logging.info(
        "Invoking onnxruntime.quantization.quantize_static with AddQDQPairToWeight=True and QuantizeBias=False.."
    )
    logging.info("Quantized model will be saved to %s.", output_model_dir)
    quantization.quantize_static(
        input_model_dir,
        output_model_dir,
        calibration_data_reader,
        extra_options=extra_options,
    )



#!/usr/bin/python

import argparse
import fnmatch
import json
import subprocess as sp
from collections import defaultdict

import pandas as pd


def _demangle(name, demangler="c++filt"):
    try:
        with sp.Popen([demangler, name], stdin=sp.PIPE, stdout=sp.PIPE) as proc:
            out, _ = proc.communicate()
            return out.decode("utf-8").strip()
    except Exception:
        return name


def _get_args():
    parser = argparse.ArgumentParser(description="onnxruntime bench tool")
    parser.add_argument("input", type=str, help="Trace input file, formatted as JSON")
    parser.add_argument(
        "--demangler",
        required=False,
        type=str,
        default="c++filt",
        help="The command to use to demangle C++ identifiers",
    )
    parser.add_argument(
        "--shape-sensitive",
        action="store_true",
        help="Perform a shape sensitive analysis of kernel execution times",
    )

    parser.add_argument(
        "--dimension-sensitive",
        action="store_true",
        help="Perform a kernel launch dimension sensitive analysis of kernel execution times",
    )

    parser.add_argument(
        "--filter",
        type=str,
        nargs="+",
        action="extend",
        help="Restrict analysis to the specified identifiers, i.e., specify a filter list. Also supports UNIX-style wildcards.",
    )
    parser.add_argument("--csv", help="Save data to csv")
    parser.add_argument("-c", "--count", type=int, default=40, help="List top N items")

    parser.add_argument(
        "--start",
        "-s",
        type=int,
        default=1,
        help="Index of the first model run to process (starting from 0, supports negative indices). "
        "Defaults to 1 to skip the first run (run 0), which is often a warmup step.",
    )
    parser.add_argument(
        "--end",
        "-e",
        type=int,
        default=None,
        help="Index of the last model run to process (exclusive, supports negative indices). "
        "Defaults to None, which means all runs starting from --start will be included.",
    )
    parser.add_argument(
        "--mapping",
        "-m",
        action="store_true",
        help="Whether dump op-kernel correlation",
    )

    args = parser.parse_args()
    return args


def _shape_to_string(shape):
    res = ""
    for dict_obj in shape:
        if len(dict_obj) > 1:
            raise ValueError("Unhandled type in _shape_to_string()")
        key = next(iter(dict_obj.keys()))
        value = next(iter(dict_obj.values()))
        if len(res) != 0:
            res += ","
        res += f"{key}({'x'.join(str(v) for v in value)})"
    return res


def _json_to_df(data, filter_matcher):
    cpu_entries = []
    gpu_entries = []

    most_recent_kernel_launch_event = None
    num_missing_kernel_launch_events = 0
    total_kernel_events = 0

    for item in data:
        cat = item.get("cat")
        if cat is None:
            continue
        dur = item.get("dur")
        if dur is None:
            continue
        arg = item.get("args")
        if arg is None:
            continue
        op_name = arg.get("op_name")

        name = item["name"]

        if not filter_matcher(name) and op_name is not None and not filter_matcher(op_name):
            continue

        if cat != "Kernel" and not name.endswith("kernel_time"):
            continue
        if name.endswith("kernel_time"):
            most_recent_kernel_launch_event = item

        block_x = arg.get("block_x", -1)
        block_y = arg.get("block_y", -1)
        block_z = arg.get("block_z", -1)
        grid_x = arg.get("grid_x", -1)
        grid_y = arg.get("grid_y", -1)
        grid_z = arg.get("grid_z", -1)

        if cat == "Kernel":
            gpu_entries.append(
                {
                    "name": name,
                    "duration": dur,
                    "dimensions": f"b{block_x}x{block_y}x{block_z},g{grid_x}x{grid_y}x{grid_z}",
                    "op_name": op_name,
                    "input_type_shape": (
                        _shape_to_string(most_recent_kernel_launch_event["args"]["input_type_shape"])
                        if most_recent_kernel_launch_event is not None
                        else "unknown"
                    ),
                }
            )
            total_kernel_events += 1
            if gpu_entries[-1]["input_type_shape"] == "unknown" and "hipMem" not in gpu_entries[-1]["name"]:
                num_missing_kernel_launch_events += 1
        else:
            cpu_entries.append(
                {
                    "name": item["args"]["op_name"],
                    "duration": dur,
                    "input_type_shape": _shape_to_string(item["args"]["input_type_shape"]),
                    "output_type_shape": _shape_to_string(item["args"]["output_type_shape"]),
                }
            )

    if num_missing_kernel_launch_events > 0:
        print(
            f"WARNING: Could not resolve shapes for {num_missing_kernel_launch_events} of {total_kernel_events} kernels."
        )

    cpu_df = pd.DataFrame(cpu_entries)
    gpu_df = pd.DataFrame(gpu_entries)
    cpu_df["count"] = 1
    gpu_df["count"] = 1
    return cpu_df, gpu_df


def _print_top_hitters(frame, args, target="cpu"):
    if len(frame) == 0:
        print(f"No {target.upper()} entries found!")
        return
    top = args.count
    group_key = ["name"]

    if target.lower() == "gpu" and args.dimension_sensitive:
        group_key.append("dimensions")

    if args.shape_sensitive:
        group_key.append("input_type_shape")

    frame2 = frame[["duration", "count"]].sum()
    frame["pct"] = 100 * (frame["duration"] / frame2["duration"])
    fields = [*group_key, "duration", "pct", "count"]
    frame1 = frame[fields].groupby(group_key).sum().reset_index()
    frame1 = frame1.sort_values(by="duration", ascending=False)[:top]
    frame1["cumulative_pct"] = frame1["pct"].cumsum()
    frame1["cumulative_dur"] = frame1["duration"].cumsum()

    if target.lower() == "gpu":
        frame1["name"] = frame1["name"].apply(lambda x: _demangle(x, args.demangler))

    print(f"\n------ Top {target.upper()} Kernel Times ------")
    print(frame1.round(2).to_string(index=False))
    if args.csv:
        frame1.to_csv(f"{args.csv}_{target}_kernel_times.csv", index=False)


def _print_op_kernel_mapping_info(cpu_df, gpu_df, num_runs, csv=None):
    # Count op occurrences in the selected runs
    op_counts = defaultdict(int)
    for op in cpu_df.T.to_dict().values():
        identifiers = (op["name"], op["input_type_shape"])
        op_counts[identifiers] += 1

    # Collect kernel stats: count/duration
    stat_dict = defaultdict(lambda: defaultdict(float))
    for kernel in gpu_df.T.to_dict().values():
        op_name = kernel["op_name"]
        if op_name is None:  # Only interested in op related kernels
            continue
        input_type_shape = kernel["input_type_shape"]
        kernel_name = kernel["name"]
        dimensions = kernel["dimensions"]
        identifiers = (op_name, input_type_shape, kernel_name, dimensions)
        stat_dict[identifiers]["count"] += 1
        stat_dict[identifiers]["duration"] += kernel["duration"]

    # Create the DataFrame for kernel entries with op correlation info
    kernel_list = []
    for identifiers, stat in stat_dict.items():
        op_name, input_type_shape, kernel_name, dimensions = identifiers
        op_count = op_counts.get((op_name, input_type_shape))
        if op_count is None:
            continue
        kernel_list.append(
            {
                "op_name": op_name,
                "input_type_shape": input_type_shape,
                "op_count": op_count / num_runs,  # Average op count per run
                "kernel_name": kernel_name,
                "kernel_dimensions": dimensions,
                "kernel_count": stat["count"] / num_runs,  # Average kernel count per run
                "kernel_avg_dur (us)": stat["duration"] / stat["count"],
                "kernel_total_dur (us)": stat["duration"] / num_runs,
            }
        )

    df = pd.DataFrame(kernel_list)
    df["op_dur (us)"] = df.groupby(["op_name", "input_type_shape"])["kernel_total_dur (us)"].transform("sum")
    df["op_avg_dur (us)"] = df["op_dur (us)"] / df["op_count"]
    df = df.sort_values(
        by=["op_dur (us)", "op_name", "input_type_shape", "kernel_total_dur (us)"],
        ascending=False,
    ).reset_index(drop=True)
    df["kernel_pct (%)"] = df["kernel_total_dur (us)"] / df["op_dur (us)"] * 100
    df["op_pct (%)"] = df["op_dur (us)"] / df["kernel_total_dur (us)"].sum() * 100
    # Move kernel_name to the end since it tends to be long
    df.insert(len(df.columns) - 1, "kernel_name", df.pop("kernel_name"))
    if csv is not None:
        df.to_csv(f"{csv}_op_kernel_mapping.csv", index=False)


def _construct_filter_matcher(args):
    if args.filter is None or len(args.filter) == 0:
        return lambda x: True
    filter_list = args.filter
    concrete_filter_set = set()
    fnmatch_filter_set = set()
    for pattern in filter_list:
        if "*" in pattern or "?" in pattern or "[" in pattern or "]" in pattern:
            fnmatch_filter_set.add(pattern)
        else:
            concrete_filter_set.add(pattern)

    def _match_item(item):
        if item in concrete_filter_set:
            return True
        return any(fnmatch.fnmatch(item, pattern) for pattern in fnmatch_filter_set)

    return _match_item


def _split_data_across_runs(data, start=1, end=None):
    """
    Splits the traces according to model runs they belong to.
    By default, we skip the first model run (run 0) and consider all subsequent runs.
    """
    # Here we assume that the traces are properly ordered, so we can simplify the splitting logic.
    model_run_splits = [i for i, item in enumerate(data) if item.get("name") == "model_run"]
    if not model_run_splits:
        print('WARNING: Could not find "model_run" event in trace. Using entire traces.')
        return data
    total_num_runs = len(model_run_splits)
    print(f"Found {total_num_runs} model_run events in trace.")

    assert -total_num_runs <= start < total_num_runs, f"Invalid start index {start}."
    if start < 0:
        start += total_num_runs
    if end is None:
        end = total_num_runs
    else:
        assert -total_num_runs <= end < total_num_runs, f"Invalid end index {end}."
        if end < 0:
            end += total_num_runs
    num_runs = end - start
    assert num_runs > 0, "No valid model runs are included in the split."
    print(f"Analyzing {num_runs} model run(s): {start}-{end - 1}.")

    # Add index 0 in case user wants to include the first model run.
    model_run_splits = [0, *model_run_splits]
    return data[model_run_splits[start] : model_run_splits[end]], num_runs


def _load_json(profile_path):
    with open(profile_path, encoding="utf-8") as file_obj:
        data = json.load(file_obj)
    if isinstance(data, dict):
        data = data["traceEvents"]
    return data


def main():
    args = _get_args()
    filter_matcher = _construct_filter_matcher(args)

    data = _load_json(args.input)
    data, num_runs = _split_data_across_runs(data, args.start, args.end)
    cpu_df, gpu_df = _json_to_df(data, filter_matcher)

    pd.set_option("display.max_colwidth", 120)
    _print_top_hitters(cpu_df, args, target="cpu")
    _print_top_hitters(gpu_df, args, target="gpu")
    if args.mapping:
        _print_op_kernel_mapping_info(cpu_df, gpu_df, num_runs, args.csv)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os

import numpy as np
import onnx

import onnxruntime as ort

model_path = "two_params_lora_model.onnx"
adapter_path = "two_params_lora_model.onnx_adapter"


def create_model(model_path: os.PathLike):
    #### Inputs
    # original input_x and its associated weight
    input_x = onnx.helper.make_tensor_value_info("input_x", onnx.TensorProto.FLOAT, [4, 4])

    # Inputs overriding default Lora initializers
    lora_param_a_input = onnx.helper.make_tensor_value_info("lora_param_a", onnx.TensorProto.FLOAT, [4, "dim"])
    lora_param_b_input = onnx.helper.make_tensor_value_info("lora_param_b", onnx.TensorProto.FLOAT, ["dim", 4])

    ### Outputs
    output = onnx.helper.make_tensor_value_info("output", onnx.TensorProto.FLOAT, [4, 4])

    #### Initializers
    # Base weight tensor proto
    weight_x = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]).reshape(4, 4).astype(np.float32)
    weight_x_tensor = onnx.helper.make_tensor("weight_x", onnx.TensorProto.FLOAT, [4, 4], weight_x.flatten())

    # tensor proto for default lora parameter A
    lora_weight_a = np.zeros([4, 0], dtype=np.float32)
    lora_weight_a_tensor = onnx.helper.make_tensor(
        "lora_param_a", onnx.TensorProto.FLOAT, [4, 0], lora_weight_a.flatten()
    )

    # tensor proto for default lora parameter B
    lora_weight_b = np.zeros([0, 4], dtype=np.float32)
    lora_weight_b_tensor = onnx.helper.make_tensor(
        "lora_param_b", onnx.TensorProto.FLOAT, [0, 4], lora_weight_b.flatten()
    )

    ##### Linear nodes
    # Create matmul for base case
    matmul_x = onnx.helper.make_node("MatMul", ["input_x", "weight_x"], ["mm_output_x"])
    # create matmul node for lora_param_a
    matmul_a = onnx.helper.make_node("MatMul", ["input_x", "lora_param_a"], ["mm_output_a"])
    # Create matmul for lora_param_b
    matmul_b = onnx.helper.make_node("MatMul", ["mm_output_a", "lora_param_b"], ["mm_output_b"])

    # Create Add
    add_node = onnx.helper.make_node("Add", ["mm_output_x", "mm_output_b"], ["output"])

    graph = onnx.helper.make_graph(
        name="two_params_lora_model",
        nodes=[matmul_x, matmul_a, matmul_b, add_node],
        inputs=[input_x, lora_param_a_input, lora_param_b_input],
        outputs=[output],
        initializer=[weight_x_tensor, lora_weight_a_tensor, lora_weight_b_tensor],
    )

    # create a model
    model = onnx.helper.make_model(graph)

    # onnx.checker.check_model(model, full_check=True)

    onnx.save_model(model, model_path)


def create_adapter(adapter_path: os.PathLike):
    """
    Creates an test adapter for the model above
    """
    param_a = np.array([3, 4, 5, 6]).astype(np.float32).reshape(4, 1)
    param_b = np.array([7, 8, 9, 10]).astype(np.float32).reshape(1, 4)
    ort_value_a = ort.OrtValue.ortvalue_from_numpy(param_a)
    ort_value_b = ort.OrtValue.ortvalue_from_numpy(param_b)

    numpy_a = ort_value_a.numpy()
    numpy_b = ort_value_b.numpy()
    np.allclose(param_a, numpy_a)
    np.allclose(param_b, numpy_b)

    print(param_a)
    print(param_b)

    name_to_value = {"lora_param_a": ort_value_a, "lora_param_b": ort_value_b}

    adapter_format = ort.AdapterFormat()
    adapter_format.set_adapter_version(1)
    adapter_format.set_model_version(1)
    adapter_format.set_parameters(name_to_value)
    adapter_format.export_adapter(adapter_path)


def read_adapter(adapter_path: os.PathLike):
    adapter = ort.AdapterFormat.read_adapter(adapter_path)
    params = adapter.get_parameters()

    assert "lora_param_a" in params
    assert "lora_param_b" in params

    numpy_a = params["lora_param_a"].numpy()
    print(numpy_a)

    numpy_b = params["lora_param_b"].numpy()
    print(numpy_b)


def run_base_model(model_path: os.PathLike):
    session = ort.InferenceSession(model_path)

    # Run the base case
    inputs = {"input_x": np.ones((4, 4), dtype=np.float32)}

    outputs = session.run(None, inputs)
    print(outputs)


def run_with_override(model_path: os.PathLike):
    session = ort.InferenceSession(model_path)

    inputs = {
        "input_x": np.ones((4, 4), dtype=np.float32),
        "lora_param_a": np.array([3, 4, 5, 6]).astype(np.float32).reshape(4, 1),
        "lora_param_b": np.array([7, 8, 9, 10]).astype(np.float32).reshape(1, 4),
    }

    outputs = session.run(None, inputs)
    print(outputs)


def run_with_adapter(model_path: os.PathLike, adapter_path: os.PathLike):
    adapter = ort.LoraAdapter()
    adapter.Load(adapter_path)

    run_options = ort.RunOptions()
    run_options.set_adapter_active(adapter)

    session = ort.InferenceSession(model_path)

    inputs = {"input_x": np.ones((4, 4), dtype=np.float32)}

    outputs = session.run(None, inputs, run_options)

    print(outputs)


if __name__ == "__main__":
    # create_model(model_path)
    # run_base_model(model_path)
    run_with_override(model_path)
    # create_adapter(adapter_path)
    # read_adapter(adapter_path)
    run_with_adapter(model_path, adapter_path)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from __future__ import annotations

import onnx

from ..onnx_model import ONNXModel
from .fusion import Fusion


class FusionGelu(Fusion):
    def __init__(self, model: ONNXModel):
        super().__init__(model, "Gelu", "Erf")

    def fuse(
        self,
        erf_node: onnx.NodeProto,
        input_name_to_nodes: dict[str, list[onnx.NodeProto]],
        output_name_to_node: dict[str, onnx.NodeProto],
    ):
        """
        Interface function that tries to fuse a node sequence containing an Erf node into a single
        Gelu node.
        """
        if (
            self.fuse_1(erf_node, input_name_to_nodes, output_name_to_node)
            or self.fuse_2(erf_node, input_name_to_nodes, output_name_to_node)
            or self.fuse_3(erf_node, input_name_to_nodes, output_name_to_node)
        ):
            self.model.set_opset_import("com.microsoft", 1)

    def fuse_1(
        self,
        erf_node: onnx.NodeProto,
        input_name_to_nodes: dict[str, list[onnx.NodeProto]],
        output_name_to_node: dict[str, onnx.NodeProto],
    ) -> bool:
        """
        This pattern is from PyTorch model
        Fuse Gelu with Erf into one node:
        Pattern 1:
                       +-------Mul(0.5)---------------------+
                       |                                    |
                       |                                    v
                    [root] --> Div -----> Erf  --> Add --> Mul -->
                              (B=1.4142...)       (1)

        Pattern 2:
                       +------------------------------------+
                       |                                    |
                       |                                    v
                    [root] --> Div -----> Erf  --> Add --> Mul -->Mul -->
                              (B=1.4142...)       (1)            (0.5)

        Note that constant input for Add and Mul could be first or second input: like either A=0.5 or B=0.5 is fine.
        """
        if erf_node.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[erf_node.output[0]]
        if len(children) != 1 or children[0].op_type != "Add":
            return False
        add_after_erf = children[0]

        if not self.has_constant_input(add_after_erf, 1):
            return False

        if add_after_erf.output[0] not in input_name_to_nodes:
            return False

        children = input_name_to_nodes[add_after_erf.output[0]]
        if len(children) != 1 or children[0].op_type != "Mul":
            return False

        mul_after_erf = children[0]

        div = self.match_parent(erf_node, "Div", 0, output_name_to_node)
        if div is None:
            return False

        if self.find_constant_input(div, 1.4142, delta=0.001) != 1:
            return False

        subgraph_input = div.input[0]

        another = 1 if mul_after_erf.input[0] == add_after_erf.output[0] else 0
        if subgraph_input == mul_after_erf.input[another]:  # pattern 2
            children = input_name_to_nodes[mul_after_erf.output[0]]
            if len(children) != 1 or children[0].op_type != "Mul":
                return False
            mul_half = children[0]
            if not self.has_constant_input(mul_half, 0.5):
                return False
            subgraph_output = mul_half.output[0]
        else:  # pattern 1
            mul_half = self.match_parent(mul_after_erf, "Mul", another, output_name_to_node)
            if mul_half is None:
                return False

            if not self.has_constant_input(mul_half, 0.5):
                return False

            if subgraph_input not in mul_half.input:
                return False

            subgraph_output = mul_after_erf.output[0]

        subgraph_nodes = [div, erf_node, add_after_erf, mul_after_erf, mul_half]
        if not self.is_safe_to_fuse_nodes(subgraph_nodes, [subgraph_output], input_name_to_nodes, output_name_to_node):
            return False

        self.nodes_to_remove.extend(subgraph_nodes)
        fused_node = onnx.helper.make_node(
            "Gelu", name=self.create_unique_node_name(), inputs=[subgraph_input], outputs=[subgraph_output]
        )
        fused_node.domain = "com.microsoft"
        self.nodes_to_add.append(fused_node)
        return True

    def fuse_2(
        self,
        erf_node: onnx.NodeProto,
        input_name_to_nodes: dict[str, list[onnx.NodeProto]],
        output_name_to_node: dict[str, onnx.NodeProto],
    ) -> bool:
        """
        This pattern is from Keras model
        Fuse Gelu with Erf into one node:
                       +------------------------------------------+
                       |                                          |
                       |                                          v
                    [root] --> Div -----> Erf  --> Add --> Mul -->Mul
                              (B=1.4142...)       (A=1)   (A=0.5)

        Note that constant input for Add and Mul could be first or second input: like either A=0.5 or B=0.5 is fine.
        """
        if erf_node.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[erf_node.output[0]]
        if len(children) != 1 or children[0].op_type != "Add":
            return False
        add_after_erf = children[0]

        if not self.has_constant_input(add_after_erf, 1):
            return False

        if add_after_erf.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[add_after_erf.output[0]]
        if len(children) != 1 or children[0].op_type != "Mul":
            return False
        mul_after_erf = children[0]

        if not self.has_constant_input(mul_after_erf, 0.5):
            return False

        if mul_after_erf.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[mul_after_erf.output[0]]
        if len(children) != 1 or children[0].op_type != "Mul":
            return False
        mul = children[0]

        div = self.match_parent(erf_node, "Div", 0, output_name_to_node)
        if div is None:
            return False

        sqrt_node = None
        if self.find_constant_input(div, 1.4142, delta=0.001) != 1:
            sqrt_node = self.match_parent(div, "Sqrt", 1, output_name_to_node)
            if sqrt_node is None:
                return False
            if not self.has_constant_input(sqrt_node, 2.0):
                return False

        subgraph_input = div.input[0]

        if subgraph_input not in mul.input:
            return False

        subgraph_nodes = [div, erf_node, add_after_erf, mul_after_erf, mul]
        if sqrt_node:
            subgraph_nodes.append(sqrt_node)

        if not self.is_safe_to_fuse_nodes(subgraph_nodes, [mul.output[0]], input_name_to_nodes, output_name_to_node):
            return False

        self.nodes_to_remove.extend(subgraph_nodes)
        fused_node = onnx.helper.make_node(
            "Gelu", name=self.create_unique_node_name(), inputs=[subgraph_input], outputs=[mul.output[0]]
        )
        fused_node.domain = "com.microsoft"
        self.nodes_to_add.append(fused_node)
        return True

    def fuse_3(
        self,
        erf_node: onnx.NodeProto,
        input_name_to_nodes: dict[str, list[onnx.NodeProto]],
        output_name_to_node: dict[str, onnx.NodeProto],
    ) -> bool:
        """
        This pattern is from TensorFlow model
        Fuse Gelu with Erf into one node:
                       +----------------------------------------------+
                       |                                              |
                       |                                              v
                    [root] --> Mul -----> Erf    -->   Add --> Mul -->Mul
                               (A=0.7071067690849304)  (B=1)  (B=0.5)

        Note that constant input for Add and Mul could be first or second input: like either A=0.5 or B=0.5 is fine.
        """

        if erf_node.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[erf_node.output[0]]
        if len(children) != 1 or children[0].op_type != "Add":
            return False
        add_after_erf = children[0]

        if not self.has_constant_input(add_after_erf, 1):
            return False

        if add_after_erf.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[add_after_erf.output[0]]
        if len(children) != 1 or children[0].op_type != "Mul":
            return False
        mul_half = children[0]

        if not self.has_constant_input(mul_half, 0.5):
            return False

        first_mul = self.match_parent(erf_node, "Mul", 0, output_name_to_node)
        if first_mul is None:
            return False

        i = self.find_constant_input(first_mul, 0.7071067690849304, delta=0.001)
        if i < 0:
            return False

        root_input_index = 1 - i
        subgraph_input = first_mul.input[root_input_index]

        if mul_half.output[0] not in input_name_to_nodes:
            return False
        children = input_name_to_nodes[mul_half.output[0]]
        if len(children) != 1 or children[0].op_type != "Mul":
            return False
        last_mul = children[0]

        if not (last_mul.input[0] == subgraph_input or last_mul.input[1] == subgraph_input):
            return False

        subgraph_nodes = [first_mul, erf_node, add_after_erf, mul_half, last_mul]
        if not self.is_safe_to_fuse_nodes(
            subgraph_nodes,
            [last_mul.output[0]],
            input_name_to_nodes,
            output_name_to_node,
        ):
            return False

        self.nodes_to_remove.extend(subgraph_nodes)
        fused_node = onnx.helper.make_node(
            "Gelu", name=self.create_unique_node_name(), inputs=[subgraph_input], outputs=[last_mul.output[0]]
        )
        fused_node.domain = "com.microsoft"
        self.nodes_to_add.append(fused_node)
        return True



#!/usr/bin/env python
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import argparse
import io  # noqa: F401
import os
import pathlib
from collections import defaultdict

import onnxruntime.capi.onnxruntime_pybind11_state as rtpy


def format_version_range(v):
    if v[1] >= 2147483647:
        return str(v[0]) + "+"
    else:
        if v[0] == v[1]:
            return str(v[0])
        else:
            return "[" + str(v[0]) + ", " + str(v[1]) + "]"


def format_type_constraints(tc):
    tcstr = ""
    firsttcitem = True
    for tcitem in tc:
        if firsttcitem:
            firsttcitem = False
        else:
            tcstr += ", "
        tcstr += tcitem
    return tcstr


def format_param_strings(params):
    firstparam = True
    s = ""
    if params:
        for param in sorted(params):
            if firstparam:
                firstparam = False
            else:
                s += "<br><br>or<br><br>"
            s += param
    return s


def expand_providers(provider_filter: [str]):
    providers = set()
    if provider_filter:
        for provider in provider_filter:
            p = provider.lower()
            if not p.endswith("executionprovider"):
                p += "executionprovider"
            providers.add(p)

    return providers


def main(output_path: pathlib.Path, provider_filter: [str]):
    providers = expand_providers(provider_filter)

    with open(output_path, "w", newline="", encoding="utf-8") as fout:
        fout.write("## Supported Operators and Data Types\n")
        fout.write(
            "*This file is automatically generated from the registered kernels by "
            "[this script](https://github.com/microsoft/onnxruntime/blob/main/tools/python/gen_opkernel_doc.py).\n"
            "Do not modify directly.*\n\n"
        )
        opdef = rtpy.get_all_operator_schema()
        paramdict = {}
        for schema in opdef:
            inputs = schema.inputs
            domain = schema.domain
            if not domain:
                domain = "ai.onnx"
            fullname = domain + "." + schema.name
            paramstr = ""
            firstinput = True
            if inputs:
                for inp in inputs:
                    if firstinput:
                        firstinput = False
                    else:
                        paramstr += "<br> "
                    paramstr += f"*in* {inp.name}:**{inp.typeStr}**"

            outputs = schema.outputs
            if outputs:
                for outp in outputs:
                    if firstinput:
                        firstinput = False
                    else:
                        paramstr += "<br> "
                    paramstr += f"*out* {outp.name}:**{outp.typeStr}**"

            paramstr += ""
            paramset = paramdict.get(fullname)
            if paramset is None:
                paramdict[fullname] = set()

            paramdict[fullname].add(paramstr)

        index = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        for op in rtpy.get_all_opkernel_def():
            domain = op.domain
            if not domain:
                domain = "ai.onnx"
            index[op.provider][domain][op.op_name].append(op)

        # TOC
        fout.write("## Execution Providers\n\n")
        for provider in sorted(index.keys()):
            if providers and provider.lower() not in providers:
                continue
            fout.write(f"- [{provider}](#{provider.lower()})\n")
        fout.write("\n---------------")

        for provider, domainmap in sorted(index.items()):
            if providers and provider.lower() not in providers:
                continue

            fout.write(f'\n\n<a name="{provider.lower()}"/>\n\n')
            fout.write(f"## Operators implemented by {provider}\n\n")
            fout.write("| Op Name | Parameters | OpSet Version | Types Supported |\n")
            fout.write("|---------|------------|---------------|-----------------|\n")
            for domain, namemap in sorted(domainmap.items()):
                fout.write("|**Operator Domain:** *" + domain + "*||||\n")
                for name, ops in sorted(namemap.items()):
                    version_type_index = defaultdict(lambda: defaultdict(set))
                    for op in ops:
                        for tname, tclist in op.type_constraints.items():
                            for c in tclist:
                                version_type_index[op.version_range][tname].add(c)

                    namefirsttime = True
                    for version_range, typemap in sorted(version_type_index.items(), key=lambda x: x[0], reverse=True):
                        if namefirsttime:
                            params = paramdict.get(domain + "." + name, None)
                            fout.write("|" + name + "|" + format_param_strings(params) + "|")
                            namefirsttime = False
                        else:
                            fout.write("|||")
                        fout.write(format_version_range(version_range) + "|")
                        for i, (tname, tcset) in enumerate(sorted(typemap.items())):
                            tnameindex = i + 1
                            tclist = []
                            for tc in sorted(tcset):
                                tclist.append(tc)
                            fout.write("**" + tname + "** = " + format_type_constraints(tclist))
                            if tnameindex < len(typemap):
                                fout.write("<br/> ")
                        fout.write("|\n")

                fout.write("| |\n| |\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="ONNX Runtime Operator Kernel Documentation Generator")
    parser.add_argument(
        "--providers",
        nargs="+",
        help="Filter to specified execution providers. Case-insensitive. "
        "Matches provider names from <ORT>/include/onnxruntime/core/graph/constants.h'. "
        "'ExecutionProvider' is automatically appended as needed. "
        "e.g. `--providers cpu cuda` will match CPUExecutionProvider and CUDAExecutionProvider.",
    )
    parser.add_argument(
        "--output_path",
        help="output markdown file path",
        type=pathlib.Path,
        required=True,
        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "OperatorKernels.md"),
    )
    args = parser.parse_args()

    main(args.output_path, args.providers)



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import unittest
from importlib.util import find_spec

import numpy as np
import numpy.typing as npt

quant_enums = {"FP4": 0, "NF4": 1}


def quantize_block_fp4(block: npt.ArrayLike):
    # quantize a block of float32 values to uint8 by simulating a binary search using pivots
    # could have used (block[:,None] - quant_map).argmin(axis=1) but there are some mismatches due to
    # floating point precision
    # block: 1-D array of normalized [-1,1] float32 values, len(block) % 2 == 0

    # pivots to find the quantization index
    # only half of the pivots are needed since the other half is symmetric
    pivots = np.array(
        [0.00260417, 0.0859375, 0.20833333, 0.29166667, 0.4166667, 0.583333, 0.8333333, 1], dtype=np.float32
    )
    # indices are not 0,1,2,3,4,5,6,7 because it is a floating point data type
    pivot_indices = np.array([0, 1, 6, 7, 4, 5, 2, 3], dtype=np.uint8)

    # signs of the block
    signs = (block < 0).astype(np.uint8) * 8

    # find the uint8 quantization index
    # argmax finds the first occurrence of True
    quant_indices = pivot_indices[(np.abs(block)[:, None] <= pivots).argmax(axis=1)] + signs

    return np.bitwise_or(np.left_shift(quant_indices[::2], 4), quant_indices[1::2])


def quantize_block_nf4(block: npt.ArrayLike):
    pivots = np.array(
        [
            -0.8480964004993439,
            -0.6106329262256622,
            -0.4599952697753906,
            -0.33967943489551544,
            -0.23460740596055984,
            -0.13791173323988914,
            -0.045525018125772476,
            0.03979014977812767,
            0.1202552504837513,
            0.2035212516784668,
            0.2920137718319893,
            0.3893125355243683,
            0.5016634166240692,
            0.6427869200706482,
            0.8614784181118011,
            1.0,
        ],
        dtype=np.float32,
    )

    quant_indices = (block[:, None] <= pivots).argmax(axis=1).astype(np.uint8)

    return np.bitwise_or(np.left_shift(quant_indices[::2], 4), quant_indices[1::2])


def quantize_blockwise_bnb4_ref(matrix_float: npt.ArrayLike, block_size: int, quant_type: str, target=None):
    if len(matrix_float.shape) != 2:
        raise ValueError("Current bnb4 block quantization only supports 2D tensors!")

    numel = matrix_float.size
    num_blocks = (numel + block_size - 1) // block_size
    quantized_numel = (numel + 1) // 2

    packed = np.zeros(quantized_numel, dtype=np.uint8)
    absmax = np.zeros(num_blocks, dtype=matrix_float.dtype)

    flattened_matrix_float = matrix_float.flatten()
    for block_idx in range(num_blocks):
        block_len = min(block_size, numel - block_idx * block_size)
        block = np.float32(flattened_matrix_float[block_idx * block_size : block_idx * block_size + block_len])

        block_absmax = np.max(np.abs(block))
        reciprocal_absmax = 1.0 / block_absmax if block_absmax != 0 else 0.0
        absmax[block_idx] = block_absmax

        if block_len % 2 != 0:
            block = np.append(block, np.float32(0.0))
            block_len += 1

        block *= reciprocal_absmax
        start = block_idx * block_size // 2
        end = start + block_len // 2
        if quant_type == "FP4":
            packed[start:end] = quantize_block_fp4(block)
        else:
            packed[start:end] = quantize_block_nf4(block)

    return (packed, absmax)


def quantize_blockwise_bnb4_target(matrix_float: npt.ArrayLike, block_size: int, quant_type: str):
    if len(matrix_float.shape) != 2:
        raise ValueError("Current int4 block quantization only supports 2D tensors!")
    quant_type_enum = quant_enums[quant_type]

    n, k = matrix_float.shape  # already transposed
    numel = n * k
    num_blocks = (numel + block_size - 1) // block_size
    quantized_numel = (numel + 1) // 2

    packed = np.zeros(quantized_numel, dtype="uint8")
    absmax = np.zeros(num_blocks, dtype=matrix_float.dtype)
    from onnxruntime.capi._pybind_state import quantize_matmul_bnb4

    quantize_matmul_bnb4(packed, matrix_float, absmax, block_size, quant_type_enum, n, k)
    return (packed, absmax)


class TestQuantizeBlockwiseBnb4(unittest.TestCase):
    @unittest.skipIf(
        find_spec("onnxruntime.training"), "Skip because training package doesn't has quantize_matmul_bnb4"
    )
    def test_quantize_blockwise_bnb4(self):
        for quant_type in ["FP4", "NF4"]:
            for k, n in [(128, 128), (32, 128), (128, 32), (52, 128), (128, 52), (73, 123)]:
                for block_size in [16, 32, 64, 128]:
                    for type in [np.float32, np.float16]:
                        matrix_float = np.random.uniform(-1, 1, (k, n)).astype(type)
                        quant_value_ref, absmax_ref = quantize_blockwise_bnb4_ref(matrix_float, block_size, quant_type)
                        quant_value, absmax = quantize_blockwise_bnb4_target(matrix_float, block_size, quant_type)
                        np.testing.assert_allclose(quant_value_ref, quant_value)
                        np.testing.assert_allclose(absmax_ref, absmax)


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

# Use triton AoT compiler to convert sparse_attention_triton.py to C source files including cubin and dispatcher.
# Example to use this script (Tested with Python 3.10 and CUDA 12.3 in Ubuntu 20.04):
#    python3 -m pip install numpy==1.26.4 torch==2.3.0 triton==2.3.0
#    python3 compile_sparse_attention.py | sh
#
# Note that sparse_attention_v1_*.cc and sparse_attention_dispatcher_*.h are the generated files.

import math
from itertools import product

import torch


def generate_triton_compile_shell_script(sm, dtype="fp16"):
    assert dtype in ["fp16", "bf16"]
    print("export TRITON_ROOT=$(pip show triton | grep Location | cut -d' ' -f2)")

    # Modify the compile.py to use custom template file template_h.txt and template_c.txt in current directory.
    # Also pass block_m to the template.
    print(
        'python -c "'
        "import sys;lines=sys.stdin.read();"
        "lines=lines.replace('template_path = Path(__file__).parent / f\\\"compile.{ext}\\\"','template_path = f\\\"compile_template_kernel_{ext}.txt\\\"');"
        'lines=lines.replace(\'\\"_placeholder\\": \\"\\",\', \'\\"_placeholder\\": \\"\\",\\n        \\"block_m\\": list(constants.values())[0],\');'
        'print(lines)"'
        "< ${TRITON_ROOT}/triton/tools/compile.py > compile.py"
    )

    out_dir = f"trition_cubin_{dtype}"
    print(f"rm -rf {out_dir}")
    print(f"mkdir -p {out_dir}")

    block_n_values = [64]
    block_d_values = [64]
    num_block_d_values = [2]
    even_n_values = [True, False]
    # Use triton compiler to compile the kernel of different combinations of constant parameters.
    for block_n, block_d, num_blocks_d, even_n in product(
        block_n_values, block_d_values, num_block_d_values, even_n_values
    ):
        head_size = block_d * num_blocks_d
        block_m = block_n
        even_m = even_n
        scalar_params = "i32,i32,i32,fp32,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32:16,i32,i32,i32"
        sig = f"*{dtype}:16,*{dtype}:16,*{dtype}:16,*{dtype}:16,*i32:16,*i32:16,{scalar_params},{block_m},{int(even_m)},{block_n},{int(even_n)},{block_d},{num_blocks_d}"
        prefix = "python compile.py sparse_attention_triton.py"
        filename = f"sparse_attention_v1_{dtype}_d{head_size}_n{block_n}_e{int(even_n)}_sm{sm}"
        name = f"sparse_attention_{dtype}_sm{sm}"
        num_warps = max(1, 2 ** int(math.log2(min(block_m, block_n, block_d) / 16)))

        # Shared memory is 96KB for V100 (sm70), 64KB for T4 (sm75), 164KB for A100 (sm80), 228KB for H100 (sm90).
        # Adjust stages so that shared memory size is within limit, and choose the one with best performance.
        sm_to_stages = {90: 3, 80: 2, 75: 2}

        num_stages = sm_to_stages[sm]

        # TODO: use different kernel name (change the name in sparse_attention_triton.py before running compile.py)
        print(
            f"{prefix} -n block_sparse_attention_kernel -o {out_dir}/{filename} --out-name {name} "
            f'-w {num_warps} -ns {num_stages} -s "{sig}" -g "(total_seq_len - past_seq_len + {block_m} - 1) / {block_m}, batch_size * num_heads, 1"'
        )

    # Generate the dispatcher.
    dispatcher = f"sparse_attention_dispatcher_{dtype}_sm{sm}"
    print(f"cd {out_dir}")
    print(f"python ${{TRITON_ROOT}}/triton/tools/link.py sparse_attention_v1_*.h -o {dispatcher}")
    print("rm *.h")

    # Remove signature hash in code.
    suffix = "0d1d2d3d4d5d678910d11d12d13d14d15d16d17d18d19d20d21d222324"
    print(f"for file in *.c; do sed -i 's/_{suffix}//g'  \"$file\"; done")

    # Recover signature hash in kernel name that is removed in previous step. Kernel name shall not be changed.
    print(
        f"for file in *.c; do sed -i 's/block_sparse_attention_kernel/block_sparse_attention_kernel_{suffix}/g'  \"$file\"; done"
    )

    # Remove signature hash from filename since we use same signature for all kernels except constants.
    # and we have constants in filename so that we can distinguish files without the hash.
    print('for file in sparse_attention_v1_*.c; do mv -- "$file" "$(echo $file | cut -f 1 -d \'.\').c"; done')

    # Change function parameters and return type. If you change the kernel interface, you will need to modify this part.
    source1 = "CUstream stream, CUdeviceptr out, CUdeviceptr Q, CUdeviceptr K, CUdeviceptr V, CUdeviceptr layout_csr_row_indices, CUdeviceptr layout_csr_col_indices, int32_t layout_csr_row_stride_h, int32_t layout_csr_col_stride_h, int32_t num_layout, float softmax_scale, int32_t stride_qb, int32_t stride_qh, int32_t stride_qm, int32_t stride_kb, int32_t stride_kh, int32_t stride_kn, int32_t stride_vb, int32_t stride_vh, int32_t stride_vn, int32_t stride_ob, int32_t stride_oh, int32_t stride_om, int32_t num_heads, int32_t num_kv_heads, int32_t total_seq_len"
    target1 = "SparseAttentionParams& params"
    source2 = "stream, out, Q, K, V, layout_csr_row_indices, layout_csr_col_indices, layout_csr_row_stride_h, layout_csr_col_stride_h, num_layout, softmax_scale, stride_qb, stride_qh, stride_qm, stride_kb, stride_kh, stride_kn, stride_vb, stride_vh, stride_vn, stride_ob, stride_oh, stride_om, num_heads, num_kv_heads, total_seq_len"
    target2 = "params"
    print(
        f"python -c \"import sys;lines=sys.stdin.read();lines=lines.replace('{source1}', '{target1}');"
        f'lines=lines.replace(\'{source2}\', \'{target2}\');print(lines)" < "{dispatcher}.c" > "{dispatcher}.h"'
    )
    print(f"sed -i 's/CUresult/Status/g'  \"{dispatcher}.h\"")

    # Remove parameter checking since we moved the validation logic to SparseAttentionParams
    print(f"sed -i '/if /d'  \"{dispatcher}.h\"")
    print(f"sed -i '/CUDA_ERROR_INVALID_VALUE/d'  \"{dispatcher}.h\"")
    print(f"sed -i '/#include/d'  \"{dispatcher}.h\"")

    print(f"rm {dispatcher}.c")

    # Use a template file to add namespace and includes to the dispatcher file.
    print(
        'python -c "'
        "from pathlib import Path;"
        "template=Path('../compile_template_dispatcher_h.txt').read_text();"
        f"code=Path('{dispatcher}.h').read_text();"
        "text=template.replace('PLACEHOLDER', code); print(text)\" "
        f"> ../{dispatcher}.h"
    )
    # rename *.c to *.cc
    print('for file in *.c; do mv -- "$file" "${file%.c}.cc"; done')

    # Move kernel files to parent directory. This might overwrite existing files in repository.
    print("mv sparse_attention_v1_* ../")

    # Clean up
    print("cd ..")
    print("rm compile.py")
    print(f"rm -rf {out_dir}")

    print("echo Done")


if __name__ == "__main__":
    major, minor = torch.cuda.get_device_capability()
    print(f"echo Generate sparse attention v1 kernels for compute capability:{major}.{minor}")
    assert major >= 7, "triton only supports compute capability >= 7.0"

    sm = major * 10 + minor
    for dtype in ["fp16", "bf16"] if major >= 8 else ["fp16"]:
        generate_triton_compile_shell_script(sm, dtype)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------
#
# This script converts stable diffusion onnx models from float to half (mixed) precision for GPU inference.
#
# Before running this script, follow README.md to setup python environment and convert stable diffusion checkpoint
# to float32 onnx models.
#
# For example, the float32 ONNX pipeline is saved to ./sd-v1-5 directory, you can optimize and convert it to float16
# like the following:
#    python optimize_pipeline.py -i ./sd-v1-5 -o ./sd-v1-5-fp16 --float16
#
# Note that the optimizations are carried out for CUDA Execution Provider at first, other EPs may not have the support
# for the fused operators. The users could disable the operator fusion manually to workaround.

import argparse
import logging
import os
import shutil
import tempfile
from pathlib import Path

import __init__  # noqa: F401. Walk-around to run this script directly
import coloredlogs
import onnx
from fusion_options import FusionOptions
from onnx_model_clip import ClipOnnxModel
from onnx_model_mmdit import MmditOnnxModel
from onnx_model_t5 import T5OnnxModel
from onnx_model_unet import UnetOnnxModel
from onnx_model_vae import VaeOnnxModel
from optimizer import optimize_by_onnxruntime, optimize_model
from packaging import version

import onnxruntime

logger = logging.getLogger(__name__)


def has_external_data(onnx_model_path):
    original_model = onnx.load_model(str(onnx_model_path), load_external_data=False)
    for initializer in original_model.graph.initializer:
        if initializer.HasField("data_location") and initializer.data_location == onnx.TensorProto.EXTERNAL:
            return True
    return False


def is_sd_3(source_dir: Path):
    return (source_dir / "text_encoder_3").exists()


def is_sdxl(source_dir: Path):
    return (
        (source_dir / "text_encoder_2").exists()
        and not (source_dir / "text_encoder_3").exists()
        and not (source_dir / "transformer").exists()
    )


def is_flux(source_dir: Path):
    return (
        (source_dir / "text_encoder_2").exists()
        and not (source_dir / "text_encoder_3").exists()
        and (source_dir / "transformer").exists()
    )


def _classify_pipeline_type(source_dir: Path):
    # May also check _class_name in model_index.json like `StableDiffusion3Pipeline` or `FluxPipeline` etc to classify.
    if is_sd_3(source_dir):
        return "sd3"

    if is_flux(source_dir):
        return "flux"

    if is_sdxl(source_dir):
        return "sdxl"

    # sd 1.x and 2.x
    return "sd"


def _get_model_list(pipeline_type: str):
    if pipeline_type == "sd3":
        return ["text_encoder", "text_encoder_2", "text_encoder_3", "transformer", "vae_encoder", "vae_decoder"]

    if pipeline_type == "flux":
        return ["text_encoder", "text_encoder_2", "transformer", "vae_encoder", "vae_decoder"]

    if pipeline_type == "sdxl":
        return ["text_encoder", "text_encoder_2", "unet", "vae_encoder", "vae_decoder"]

    assert pipeline_type == "sd"
    return ["text_encoder", "unet", "vae_encoder", "vae_decoder"]


def _optimize_sd_pipeline(
    source_dir: Path,
    target_dir: Path,
    pipeline_type: str,
    model_list: list[str],
    use_external_data_format: bool | None,
    float16: bool,
    bfloat16: bool,
    force_fp32_ops: list[str],
    enable_runtime_optimization: bool,
    args,
):
    """Optimize onnx models used in stable diffusion onnx pipeline and optionally convert to float16.

    Args:
        source_dir (Path): Root of input directory of stable diffusion onnx pipeline with float32 models.
        target_dir (Path): Root of output directory of stable diffusion onnx pipeline with optimized models.
        model_list (List[str]): list of directory names with onnx model.
        use_external_data_format (Optional[bool]): use external data format.
        float16 (bool): use half precision
        bfloat16 (bool): use bfloat16 as fallback if float16 is also provided.
        force_fp32_ops(List[str]): operators that are forced to run in float32.
        enable_runtime_optimization(bool): run graph optimization using Onnx Runtime.

    Raises:
        RuntimeError: input onnx model does not exist
        RuntimeError: output onnx model path existed
    """
    is_flux_pipeline = pipeline_type == "flux"
    model_type_mapping = {
        "transformer": "mmdit",
        "unet": "unet",
        "vae_encoder": "vae",
        "vae_decoder": "vae",
        "text_encoder": "clip",
        "text_encoder_2": "t5" if is_flux_pipeline else "clip",
        "text_encoder_3": "t5",  # t5-v1_1-xxl is used in SD 3.x text_encoder_3 and Flux text_encoder_2.
        "safety_checker": "unet",
    }

    model_type_class_mapping = {
        "unet": UnetOnnxModel,
        "vae": VaeOnnxModel,
        "clip": ClipOnnxModel,
        "t5": T5OnnxModel,
        "mmdit": MmditOnnxModel,
    }

    force_fp32_operators = {
        "unet": [],
        "vae_encoder": [],
        "vae_decoder": [],
        "text_encoder": [],
        "text_encoder_2": [],
        "safety_checker": [],
        "text_encoder_3": [],
        "transformer": [],
    }

    # The node block list is generated by running the fp32 model and get statistics of node inputs and outputs.
    # Nodes with any input or output of float or double data type, but value ouf of range of float16 are candidates.
    #   python optimize_pipeline.py -i ./flux1_schnell_onnx/fp32 -o ./flux1_schnell_onnx/fp32_opt
    #   export ORT_DEBUG_NODE_IO_DUMP_STATISTICS_DATA=1
    #   export ORT_DEBUG_NODE_IO_DUMP_INPUT_DATA=1
    #   export ORT_DEBUG_NODE_IO_DUMP_OUTPUT_DATA=1
    #   python benchmark.py --height 1024 --width 1024 --steps 4 -b 1 -v Flux.1S -p flux1_schnell_onnx/fp32_opt -e optimum >stdout.txt 2>stderr.txt
    # Warning: The node name might change in different export settings. See benchmark_flux.sh for the settings.
    flux_node_block_list = {
        "text_encoder_2": [
            "/encoder/block.10/layer.1/DenseReluDense/wo/MatMul",
            "SkipLayerNorm_20",
            "SkipLayerNorm_21",
            "SkipLayerNorm_22",
            "SkipLayerNorm_23",
            "SkipLayerNorm_24",
            "SkipLayerNorm_25",
            "SkipLayerNorm_26",
            "SkipLayerNorm_27",
            "SkipLayerNorm_28",
            "SkipLayerNorm_29",
            "SkipLayerNorm_30",
            "SkipLayerNorm_31",
            "SkipLayerNorm_32",
            "SkipLayerNorm_33",
            "SkipLayerNorm_34",
            "SkipLayerNorm_35",
            "SkipLayerNorm_36",
            "SkipLayerNorm_37",
            "SkipLayerNorm_38",
            "SkipLayerNorm_39",
            "SkipLayerNorm_40",
            "SkipLayerNorm_41",
            "SkipLayerNorm_42",
            "SkipLayerNorm_43",
            "SkipLayerNorm_44",
            "SkipLayerNorm_45",
            "/encoder/block.23/layer.1/DenseReluDense/wo/MatMul",
            "SkipLayerNorm_46",
        ],
        "vae_decoder": [
            "/decoder/mid_block/attentions.0/MatMul",
            "/decoder/mid_block/attentions.0/Softmax",
        ],
        "transformer": [
            "/transformer_blocks.18/Mul_5",
            "/transformer_blocks.18/Add_7",
            "/Concat_1",
            "LayerNorm_76",
            "/single_transformer_blocks.0/Add",
            "LayerNorm_77",
            "/single_transformer_blocks.1/Add",
            "LayerNorm_78",
            "/single_transformer_blocks.2/Add",
            "LayerNorm_79",
            "/single_transformer_blocks.3/Add",
            "LayerNorm_80",
            "/single_transformer_blocks.4/Add",
            "LayerNorm_81",
            "/single_transformer_blocks.5/Add",
            "LayerNorm_82",
            "/single_transformer_blocks.6/Add",
            "LayerNorm_83",
            "/single_transformer_blocks.7/Add",
            "LayerNorm_84",
            "/single_transformer_blocks.8/Add",
            "LayerNorm_85",
            "/single_transformer_blocks.9/Add",
            "LayerNorm_86",
            "/single_transformer_blocks.10/Add",
            "LayerNorm_87",
            "/single_transformer_blocks.11/Add",
            "LayerNorm_88",
            "/single_transformer_blocks.12/Add",
            "LayerNorm_89",
            "/single_transformer_blocks.13/Add",
            "LayerNorm_90",
            "/single_transformer_blocks.14/Add",
            "LayerNorm_91",
            "/single_transformer_blocks.15/Add",
            "LayerNorm_92",
            "/single_transformer_blocks.16/Add",
            "LayerNorm_93",
            "/single_transformer_blocks.17/Add",
            "LayerNorm_94",
            "/single_transformer_blocks.18/Add",
            "LayerNorm_95",
            "/single_transformer_blocks.19/Add",
            "LayerNorm_96",
            "/single_transformer_blocks.20/Add",
            "LayerNorm_97",
            "/single_transformer_blocks.21/Add",
            "LayerNorm_98",
            "/single_transformer_blocks.22/Add",
            "LayerNorm_99",
            "/single_transformer_blocks.23/Add",
            "LayerNorm_100",
            "/single_transformer_blocks.24/Add",
            "LayerNorm_101",
            "/single_transformer_blocks.25/Add",
            "LayerNorm_102",
            "/single_transformer_blocks.26/Add",
            "LayerNorm_103",
            "/single_transformer_blocks.27/Add",
            "LayerNorm_104",
            "/single_transformer_blocks.28/Add",
            "LayerNorm_105",
            "/single_transformer_blocks.29/Add",
            "LayerNorm_106",
            "/single_transformer_blocks.30/Add",
            "LayerNorm_107",
            "/single_transformer_blocks.31/Add",
            "LayerNorm_108",
            "/single_transformer_blocks.32/Add",
            "LayerNorm_109",
            "/single_transformer_blocks.33/Add",
            "LayerNorm_110",
            "/single_transformer_blocks.34/Add",
            "LayerNorm_111",
            "/single_transformer_blocks.35/Add",
            "LayerNorm_112",
            "/single_transformer_blocks.36/Add",
            "LayerNorm_113",
            "/single_transformer_blocks.37/Add",
            "/Shape",
            "/Slice",
        ],
    }

    sd3_node_block_list = {"text_encoder_3": flux_node_block_list["text_encoder_2"]}

    if force_fp32_ops:
        for fp32_operator in force_fp32_ops:
            parts = fp32_operator.split(":")
            if len(parts) == 2 and parts[0] in force_fp32_operators and (parts[1] and parts[1][0].isupper()):
                force_fp32_operators[parts[0]].append(parts[1])
            else:
                raise ValueError(
                    f"--force_fp32_ops shall be in the format of module:operator like unet:Attention, got {fp32_operator}"
                )

    op_counters = {}
    for name, model_type in model_type_mapping.items():
        onnx_model_path = source_dir / name / "model.onnx"
        if not os.path.exists(onnx_model_path):
            if name != "safety_checker" and name in model_list:
                logger.warning("input onnx model does not exist: %s", onnx_model_path)
            # some model are optional so we do not raise error here.
            continue

        # Prepare output directory
        optimized_model_path = target_dir / name / "model.onnx"
        if os.path.exists(optimized_model_path):
            if not args.overwrite:
                logger.warning("Skipped optimization since the target file existed: %s", optimized_model_path)
            continue
        output_dir = optimized_model_path.parent
        output_dir.mkdir(parents=True, exist_ok=True)

        if use_external_data_format is None:
            use_external_data_format = has_external_data(onnx_model_path)

        # Graph fusion before fp16 conversion, otherwise they cannot be fused later.
        logger.info("Optimize %s ...", onnx_model_path)

        args.model_type = model_type
        fusion_options = FusionOptions.parse(args)

        if model_type in ["unet"]:
            # Some optimizations are not available in v1.14 or older version: packed QKV and BiasAdd
            has_all_optimizations = version.parse(onnxruntime.__version__) >= version.parse("1.15.0")
            fusion_options.enable_packed_kv = float16 and fusion_options.enable_packed_kv
            fusion_options.enable_packed_qkv = float16 and has_all_optimizations and fusion_options.enable_packed_qkv
            fusion_options.enable_bias_add = has_all_optimizations and fusion_options.enable_bias_add

        m = optimize_model(
            str(onnx_model_path),
            model_type=model_type,
            num_heads=0,  # will be deduced from graph
            hidden_size=0,  # will be deduced from graph
            opt_level=0,
            optimization_options=fusion_options,
            use_gpu=True,
            provider=args.provider,
        )

        if float16:
            model_node_block_list = (
                flux_node_block_list if is_flux_pipeline else sd3_node_block_list if pipeline_type == "sd3" else {}
            )
            if name in model_node_block_list:
                # Opset 12 does not support bfloat16.
                # By default, optimum exports T5 model with opset 12. So we need to check the opset version.
                use_bfloat16 = bfloat16
                if use_bfloat16:
                    for opset in m.model.opset_import:
                        if opset.domain in ["", "ai.onnx"] and opset.version < 13:
                            logger.warning(
                                "onnx model requires opset 13 or higher to use bfloat16. Fall back to float32."
                            )
                            use_bfloat16 = False

                m.convert_float_to_float16(
                    keep_io_types=False,
                    node_block_list=model_node_block_list[name],
                    use_bfloat16_as_blocked_nodes_dtype=use_bfloat16,
                )
            # For SD-XL, use FP16 in VAE decoder will cause NaN and black image so we keep it in FP32.
            elif pipeline_type in ["sdxl"] and name in ["vae_decoder"]:
                logger.info("Skip converting %s to float16 to avoid NaN", name)
            else:
                logger.info("Convert %s to float16 ...", name)
                m.convert_float_to_float16(
                    keep_io_types=False,
                    op_block_list=force_fp32_operators[name],
                )

        if enable_runtime_optimization:
            # Use this step to see the final graph that executed by Onnx Runtime.
            with tempfile.TemporaryDirectory() as tmp_dir:
                # Save to a temporary file so that we can load it with Onnx Runtime.
                logger.info("Saving a temporary model to run OnnxRuntime graph optimizations...")
                tmp_model_path = Path(tmp_dir) / "model.onnx"
                m.save_model_to_file(str(tmp_model_path), use_external_data_format=use_external_data_format)
                ort_optimized_model_path = Path(tmp_dir) / "optimized.onnx"
                optimize_by_onnxruntime(
                    str(tmp_model_path),
                    use_gpu=True,
                    provider=args.provider,
                    optimized_model_path=str(ort_optimized_model_path),
                    save_as_external_data=use_external_data_format,
                )
                model = onnx.load(str(ort_optimized_model_path), load_external_data=True)
                m = model_type_class_mapping[model_type](model)

        m.get_operator_statistics()
        op_counters[name] = m.get_fused_operator_statistics()
        m.save_model_to_file(str(optimized_model_path), use_external_data_format=use_external_data_format)
        logger.info("%s is optimized", name)
        logger.info("*" * 20)

    return op_counters


def _copy_extra_directory(source_dir: Path, target_dir: Path, model_list: list[str]):
    """Copy extra directory that does not have onnx model

    Args:
        source_dir (Path): source directory
        target_dir (Path): target directory
        model_list (List[str]): list of directory names with onnx model.

    Raises:
        RuntimeError: source path does not exist
    """
    extra_dirs = ["scheduler", "tokenizer", "tokenizer_2", "tokenizer_3", "feature_extractor"]

    for name in extra_dirs:
        source_path = source_dir / name
        if not os.path.exists(source_path):
            continue

        target_path = target_dir / name
        if target_path.exists():
            shutil.rmtree(target_path)
        shutil.copytree(source_path, target_path)
        logger.info("%s => %s", source_path, target_path)

    extra_files = ["model_index.json"]
    for name in extra_files:
        source_path = source_dir / name
        if not os.path.exists(source_path):
            raise RuntimeError(f"source path does not exist: {source_path}")

        target_path = target_dir / name
        shutil.copyfile(source_path, target_path)
        logger.info("%s => %s", source_path, target_path)

    # Some directory are optional
    for onnx_model_dir in model_list:
        source_path = source_dir / onnx_model_dir / "config.json"
        target_path = target_dir / onnx_model_dir / "config.json"
        if source_path.exists():
            target_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(source_path, target_path)
            logger.info("%s => %s", source_path, target_path)


def optimize_stable_diffusion_pipeline(
    input_dir: str,
    output_dir: str,
    overwrite: bool,
    use_external_data_format: bool | None,
    float16: bool,
    enable_runtime_optimization: bool,
    args,
):
    if os.path.exists(output_dir):
        if overwrite:
            shutil.rmtree(output_dir, ignore_errors=True)

    source_dir = Path(input_dir)
    target_dir = Path(output_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    pipeline_type = _classify_pipeline_type(source_dir)
    model_list = _get_model_list(pipeline_type)

    _copy_extra_directory(source_dir, target_dir, model_list)

    return _optimize_sd_pipeline(
        source_dir,
        target_dir,
        pipeline_type,
        model_list,
        use_external_data_format,
        float16,
        args.bfloat16,
        args.force_fp32_ops,
        enable_runtime_optimization,
        args,
    )


def parse_arguments(argv: list[str] | None = None):
    """Parse arguments

    Returns:
        Namespace: arguments
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-i",
        "--input",
        required=True,
        type=str,
        help="Root of input directory of stable diffusion onnx pipeline with float32 models.",
    )

    parser.add_argument(
        "-o",
        "--output",
        required=True,
        type=str,
        help="Root of output directory of stable diffusion onnx pipeline with optimized models.",
    )

    parser.add_argument(
        "--float16",
        required=False,
        action="store_true",
        help="Output models of float16, except some nodes falls back to float32 or bfloat16 to avoid overflow.",
    )
    parser.set_defaults(float16=False)

    parser.add_argument(
        "--bfloat16",
        required=False,
        action="store_true",
        help="Allow bfloat16 as fallback if --float16 is also provided.",
    )
    parser.set_defaults(bfloat16=False)

    parser.add_argument(
        "--force_fp32_ops",
        required=False,
        nargs="+",
        type=str,
        help="Force given operators (like unet:Attention) to run in float32. It is case sensitive!",
    )

    parser.add_argument(
        "--inspect",
        required=False,
        action="store_true",
        help="Save the optimized graph from Onnx Runtime. "
        "This option has no impact on inference performance except it might reduce session creation time.",
    )
    parser.set_defaults(inspect=False)

    parser.add_argument(
        "--overwrite",
        required=False,
        action="store_true",
        help="Overwrite exists files.",
    )
    parser.set_defaults(overwrite=False)

    parser.add_argument(
        "-e",
        "--use_external_data_format",
        required=False,
        action="store_true",
        help="Onnx model larger than 2GB need to use external data format. "
        "If specified, save each onnx model to two files: one for onnx graph, another for weights. "
        "If not specified, use same format as original model by default. ",
    )
    parser.set_defaults(use_external_data_format=None)

    parser.add_argument(
        "--provider",
        required=False,
        type=str,
        default=None,
        help="Execution provider to use.",
    )

    FusionOptions.add_arguments(parser)

    args = parser.parse_args(argv)
    return args


def main(argv: list[str] | None = None):
    args = parse_arguments(argv)

    logger.info("Arguments: %s", str(args))

    # Return op counters for testing purpose.
    return optimize_stable_diffusion_pipeline(
        args.input, args.output, args.overwrite, args.use_external_data_format, args.float16, args.inspect, args
    )


if __name__ == "__main__":
    coloredlogs.install(fmt="%(funcName)20s: %(message)s")
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os

import onnx

from onnxruntime.capi._pybind_state import get_model_after_loading_checkpoint as _load_checkpoint_to_model
from onnxruntime.capi._pybind_state import save_checkpoint as _save_checkpoint


def save_checkpoint(
    parameters: tuple[list[onnx.TensorProto], list[onnx.TensorProto]],
    path_to_checkpoint: str | os.PathLike,
    nominal_checkpoint: bool = False,
) -> None:
    """Saves the parameters to the checkpoint directory path_to_checkpoint.

    Args:
        parameters tuple(trainable_params, non_trainable_params): The parameters to save to the checkpoint file.
        path_to_checkpoint: The path to the checkpoint directory.
        nominal_checkpoint: If True, the checkpoint is saved as a nominal checkpoint. Default is False.
    """

    if parameters is None:
        raise RuntimeError("No checkpoint parameters provided.")

    trainable_params, non_trainable_params = parameters
    trainable_params = [param.SerializeToString() for param in trainable_params]
    non_trainable_params = [param.SerializeToString() for param in non_trainable_params]
    _save_checkpoint(trainable_params, non_trainable_params, os.fspath(path_to_checkpoint), nominal_checkpoint)


def load_checkpoint_to_model(path_to_checkpoint: str | os.PathLike, model: onnx.ModelProto) -> None:
    """Loads the checkpoint to an onnx inference model.

    Args:
        path_to_checkpoint (str): The path to the checkpoint directory.
        model (onnx.ModelProto): The model to load the checkpoint to.
    """

    model.ParseFromString(_load_checkpoint_to_model(os.fspath(path_to_checkpoint), model.SerializeToString()))



#!/usr/bin/env python

# This file is copied and adapted from https://github.com/onnx/onnx repository.
# There was no copyright statement on the file at the time of copying.

import argparse
import os
import pathlib
import sys
from collections import defaultdict

import numpy as np

import onnxruntime.capi.onnxruntime_pybind11_state as rtpy
from onnxruntime.capi.onnxruntime_pybind11_state import schemadef  # noqa: F401
from onnxruntime.capi.onnxruntime_pybind11_state.schemadef import OpSchema

ONNX_ML = not bool(os.getenv("ONNX_ML") == "0")
ONNX_DOMAIN = "onnx"
ONNX_ML_DOMAIN = "onnx-ml"

if ONNX_ML:
    ext = "-ml.md"
else:
    ext = ".md"


def display_number(v):  # type: (int) -> Text
    if OpSchema.is_infinite(v):
        return "&#8734;"
    return str(v)


def should_render_domain(domain, domain_filter):  # type: (Text) -> bool
    if domain in (ONNX_DOMAIN, ONNX_ML_DOMAIN) or domain == "" or domain == "ai.onnx.ml":
        return False

    if domain_filter and domain not in domain_filter:
        return False

    return True


def format_name_with_domain(domain, schema_name):  # type: (Text, Text) -> Text
    if domain:
        return f"{domain}.{schema_name}"
    else:
        return schema_name


def format_name_with_version(schema_name, version):  # type: (Text, Text) -> Text
    return f"{schema_name}-{version}"


def display_attr_type(v):  # type: (OpSchema.AttrType) -> Text
    assert isinstance(v, OpSchema.AttrType)
    s = str(v)
    s = s[s.rfind(".") + 1 :].lower()
    if s[-1] == "s":
        s = "list of " + s
    return s


def display_domain(domain):  # type: (Text) -> Text
    if domain:
        return f"the '{domain}' operator set"
    else:
        return "the default ONNX operator set"


def display_domain_short(domain):  # type: (Text) -> Text
    if domain:
        return domain
    else:
        return "ai.onnx (default)"


def display_version_link(name, version):  # type: (Text, int) -> Text
    changelog_md = "Changelog" + ext
    name_with_ver = f"{name}-{version}"
    return f'<a href="{changelog_md}#{name_with_ver}">{name_with_ver}</a>'


def display_function_version_link(name, version):  # type: (Text, int) -> Text
    changelog_md = "FunctionsChangelog" + ext
    name_with_ver = f"{name}-{version}"
    return f'<a href="{changelog_md}#{name_with_ver}">{name_with_ver}</a>'


def get_attribute_value(attr):  # type: (AttributeProto) -> Any
    if attr.HasField("f"):
        return attr.f
    elif attr.HasField("i"):
        return attr.i
    elif attr.HasField("s"):
        return attr.s
    elif attr.HasField("t"):
        return attr.t
    elif attr.HasField("g"):
        return attr.g
    elif len(attr.floats):
        return list(attr.floats)
    elif len(attr.ints):
        return list(attr.ints)
    elif len(attr.strings):
        return list(attr.strings)
    elif len(attr.tensors):
        return list(attr.tensors)
    elif len(attr.graphs):
        return list(attr.graphs)
    else:
        raise ValueError(f"Unsupported ONNX attribute: {attr}")


def display_schema(schema, versions):  # type: (OpSchema, Sequence[OpSchema]) -> Text
    s = ""

    # doc
    schemadoc = schema.doc
    if schemadoc:
        s += "\n"
        s += "\n".join("  " + line for line in schemadoc.lstrip().splitlines())
        s += "\n"

    # since version
    s += "\n#### Version\n"
    if schema.support_level == OpSchema.SupportType.EXPERIMENTAL:
        s += "\nNo versioning maintained for experimental ops."
    else:
        s += (
            "\nThis version of the operator has been "
            + ("deprecated" if schema.deprecated else "available")
            + f" since version {schema.since_version}"
        )
        s += f" of {display_domain(schema.domain)}.\n"
        if len(versions) > 1:
            # TODO: link to the Changelog.md
            s += "\nOther versions of this operator: {}\n".format(
                ", ".join(
                    format_name_with_version(format_name_with_domain(v.domain, v.name), v.since_version)
                    for v in versions[:-1]
                )
            )

    # If this schema is deprecated, don't display any of the following sections
    if schema.deprecated:
        return s

    # attributes
    attribs = schema.attributes
    if attribs:
        s += "\n#### Attributes\n\n"
        s += "<dl>\n"
        for _, attr in sorted(attribs.items()):
            # option holds either required or default value
            opt = ""
            if attr.required:
                opt = "required"
            elif hasattr(attr, "default_value") and attr.default_value.name:
                default_value = get_attribute_value(attr.default_value)

                def format_value(value):  # type: (Any) -> Text
                    if isinstance(value, float):
                        value = np.round(value, 5)
                    if isinstance(value, (bytes, bytearray)) and sys.version_info[0] == 3:  # noqa: YTT201
                        value = value.decode("utf-8")
                    return str(value)

                if isinstance(default_value, list):
                    default_value = [format_value(val) for val in default_value]
                else:
                    default_value = format_value(default_value)
                opt = f"default is {default_value}"

            s += "<dt><tt>{}</tt> : {}{}</dt>\n".format(
                attr.name, display_attr_type(attr.type), f" ({opt})" if opt else ""
            )
            s += f"<dd>{attr.description}</dd>\n"
        s += "</dl>\n"

    # inputs
    s += "\n#### Inputs"
    if schema.min_input != schema.max_input:
        s += f" ({display_number(schema.min_input)} - {display_number(schema.max_input)})"
    s += "\n\n"

    inputs = schema.inputs
    if inputs:
        s += "<dl>\n"
        for inp in inputs:
            option_str = ""
            if OpSchema.FormalParameterOption.Optional == inp.option:
                option_str = " (optional)"
            elif OpSchema.FormalParameterOption.Variadic == inp.option:
                if inp.isHomogeneous:
                    option_str = " (variadic)"
                else:
                    option_str = " (variadic, heterogeneous)"
            s += f"<dt><tt>{inp.name}</tt>{option_str} : {inp.typeStr}</dt>\n"
            s += f"<dd>{inp.description}</dd>\n"

    s += "</dl>\n"

    # outputs
    s += "\n#### Outputs"
    if schema.min_output != schema.max_output:
        s += f" ({display_number(schema.min_output)} - {display_number(schema.max_output)})"
    s += "\n\n"
    outputs = schema.outputs
    if outputs:
        s += "<dl>\n"
        for output in outputs:
            option_str = ""
            if OpSchema.FormalParameterOption.Optional == output.option:
                option_str = " (optional)"
            elif OpSchema.FormalParameterOption.Variadic == output.option:
                if output.isHomogeneous:
                    option_str = " (variadic)"
                else:
                    option_str = " (variadic, heterogeneous)"
            s += f"<dt><tt>{output.name}</tt>{option_str} : {output.typeStr}</dt>\n"
            s += f"<dd>{output.description}</dd>\n"

    s += "</dl>\n"

    # type constraints
    s += "\n#### Type Constraints"
    s += "\n\n"
    typecons = schema.type_constraints
    if typecons:
        s += "<dl>\n"
        for type_constraint in typecons:
            allowed_types = type_constraint.allowed_type_strs
            allowed_type_str = ""
            if len(allowed_types) > 0:
                allowed_type_str = allowed_types[0]
            for allowedType in allowed_types[1:]:  # noqa: N806
                allowed_type_str += ", " + allowedType
            s += f"<dt><tt>{type_constraint.type_param_str}</tt> : {allowed_type_str}</dt>\n"
            s += f"<dd>{type_constraint.description}</dd>\n"
        s += "</dl>\n"

    return s


def display_function(function, versions, domain=ONNX_DOMAIN):  # type: (FunctionProto, List[int], Text) -> Text
    s = ""

    if domain:
        domain_prefix = f"{ONNX_ML_DOMAIN}."
    else:
        domain_prefix = ""

    # doc
    if function.doc_string:
        s += "\n"
        s += "\n".join("  " + line for line in function.doc_string.lstrip().splitlines())
        s += "\n"

    # since version
    s += "\n#### Version\n"
    s += f"\nThis version of the function has been available since version {function.since_version}"
    s += f" of {display_domain(domain_prefix)}.\n"
    if len(versions) > 1:
        s += "\nOther versions of this function: {}\n".format(
            ", ".join(
                display_function_version_link(domain_prefix + function.name, v)
                for v in versions
                if v != function.since_version
            )
        )

    # inputs
    s += "\n#### Inputs"
    s += "\n\n"
    if function.input:
        s += "<dl>\n"
        for input in function.input:
            s += f"<dt>{input}; </dt>\n"
        s += "<br/></dl>\n"

    # outputs
    s += "\n#### Outputs"
    s += "\n\n"
    if function.output:
        s += "<dl>\n"
        for output in function.output:
            s += f"<dt>{output}; </dt>\n"
        s += "<br/></dl>\n"

        # attributes
    if function.attribute:
        s += "\n#### Attributes\n\n"
        s += "<dl>\n"
        for attr in function.attribute:
            s += f"<dt>{attr};<br/></dt>\n"
        s += "</dl>\n"

    return s


def support_level_str(level):  # type: (OpSchema.SupportType) -> Text
    return "<sub>experimental</sub> " if level == OpSchema.SupportType.EXPERIMENTAL else ""


def main(output_path: str, domain_filter: [str]):
    with open(output_path, "w", newline="", encoding="utf-8") as fout:
        fout.write("## Contrib Operator Schemas\n")
        fout.write(
            "*This file is automatically generated from the registered contrib operator schemas by "
            "[this script](https://github.com/microsoft/onnxruntime/blob/main/tools/python/gen_contrib_doc.py).\n"
            "Do not modify directly.*\n"
        )

        # domain -> support level -> name -> [schema]
        index = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))  # type: Dict[Text, Dict[int, Dict[Text, List[OpSchema]]]]

        for schema in rtpy.get_all_operator_schema():
            index[schema.domain][int(schema.support_level)][schema.name].append(schema)

        fout.write("\n")

        # Preprocess the Operator Schemas
        # [(domain, [(support_level, [(schema name, current schema, all versions schemas)])])]
        operator_schemas = []  # type: List[Tuple[Text, List[Tuple[int, List[Tuple[Text, OpSchema, List[OpSchema]]]]]]]
        exsting_ops = set()  # type: Set[Text]
        for domain, _supportmap in sorted(index.items()):
            if not should_render_domain(domain, domain_filter):
                continue

            processed_supportmap = []
            for _support, _namemap in sorted(_supportmap.items()):
                processed_namemap = []
                for n, unsorted_versions in sorted(_namemap.items()):
                    versions = sorted(unsorted_versions, key=lambda s: s.since_version)
                    schema = versions[-1]
                    if schema.name in exsting_ops:
                        continue
                    exsting_ops.add(schema.name)
                    processed_namemap.append((n, schema, versions))
                processed_supportmap.append((_support, processed_namemap))
            operator_schemas.append((domain, processed_supportmap))

        # Table of contents
        for domain, supportmap in operator_schemas:
            s = f"* {display_domain_short(domain)}\n"
            fout.write(s)

            for _, namemap in supportmap:
                for n, schema, versions in namemap:  # noqa: B007
                    s = f'  * {support_level_str(schema.support_level)}<a href="#{format_name_with_domain(domain, n)}">{format_name_with_domain(domain, n)}</a>\n'
                    fout.write(s)

        fout.write("\n")

        for domain, supportmap in operator_schemas:
            s = f"## {display_domain_short(domain)}\n"
            fout.write(s)

            for _, namemap in supportmap:
                for op_type, schema, versions in namemap:
                    # op_type
                    s = (
                        '### {}<a name="{}"></a><a name="{}">**{}**'
                        + (" (deprecated)" if schema.deprecated else "")
                        + "</a>\n"
                    ).format(
                        support_level_str(schema.support_level),
                        format_name_with_domain(domain, op_type),
                        format_name_with_domain(domain, op_type.lower()),
                        format_name_with_domain(domain, op_type),
                    )

                    s += display_schema(schema, versions)

                    s += "\n\n"

                    fout.write(s)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="ONNX Runtime Contrib Operator Documentation Generator")
    parser.add_argument(
        "--domains",
        nargs="+",
        help="Filter to specified domains. e.g. `--domains com.microsoft com.microsoft.nchwc`",
    )
    parser.add_argument(
        "--output_path",
        help="output markdown file path",
        type=pathlib.Path,
        required=True,
        default=os.path.join(os.path.dirname(os.path.realpath(__file__)), "ContribOperators.md"),
    )
    args = parser.parse_args()
    output_path = args.output_path.resolve()

    main(output_path, args.domains)
